import gurobipy as gp
from gurobipy import GRB
import numpy as np
import sys
import os
import math
import time

from gurobi_modelanalyzer import common


#
#   Ill conditioning explainer.   If the model has basis statuses, it will
#   use them, computing the factorization if needed.   If no basis statuses
#   are available, solve the LP to optimality (or whatever the final status
#   is) and use that basis.
#   Computing exact Kappa can be expensive.   Set KappaExact to 0 to avoid,
#   KappaExact to 1 to force this (in addition to estimated Kappa).  Default
#   of -1 decides based on model size.
#
OFF = 0
MODERATE = 1
VERBOSE = 2
_debug = OFF  # Change to MODERATE or VERBOSE as needed
_debugger = OFF

SOLVELP = 0  # relobjtype choices
SOLVEQP = 1
SOLVEMIP = 2
BYROWS = 1  # expltype choices
BYCOLS = 2
DEFAULT = 0  # method choices.  Default = no regularization.
ANGLES = 1
LASSO = 2  # One norm regularization.
RLS = 3  # Two norm regularization.
CON = 0
VAR = 1
COEFF = 2
DEFSMALLTOL = 1e-13
COMBINEDROW = "GRB_Combined_Row"
COMBINEDCOL = "GRB_Combined_Column"


def kappa_explain(
    model,
    data=None,
    KappaExact=-1,
    prmfile=None,
    relobjtype="LP",
    expltype="ROWS",
    method="DEFAULT",
    smalltol=DEFSMALLTOL,
    submatrix=False,
):
    #
    #   Help function info
    #
    """Ill conditioning explainer.   Any basis statuses, in the model
    will be used, computing the factorization if needed.   If no statuses
    are available, solve the LP to optimality (or whatever the final
    solve status is) and use that basis.

    Arguments:
    model      (required) The LP model whose basis will be examined.
                          Basis can be from a completed solve or statuses.
                          LP will be optimized if no basis present
    data                  Do not provide.  Used only for Gurobi routines.
    KappaExact (optional) 1 = display exact condition number in stats
                          0 = display less computationally intensive estimate
                         -1 = (default) decide based on problem size.
    prmfile    (optional) Parameter settings file for the subproblem
                          solved to derive the ill conditioning certificate.
    relobjtype (optional) Type of subproblem to create for calculation of
                          ill conditioning certificate.
                          "LP" (default)
                          "QP"
    expltype   (optional) Row ("ROWS") or column ("COLS") based computation
                          and explanation. Default is by row.  One type of
                          explanation is often much smaller than the other.
    method     (optional) Alternate computation methods (may help generate
                          smaller explanation).
                          "DEFAULT" = Basic subproblem (no regularization).
                          "ANGLES"  = Simpler calculation based on inner
                                      products.  Potentially faster but only
                                      finds explanations of 2 rows or
                                      columns.
                          "LASSO"   = One norm regularization.
                          "RLS"     = Two norm regularization.
    smalltol   (optional) Tolerance below which certificate of ill
                          conditioning values are treated as zero.   If
                          left at default of 1e-13, row or column norm and
                          machine precision will be incorporated.
    submatrix  (optional) Whether to postprocess the explanation down to
                          a smaller submatrix.  Default is False.
    Returns:              For all method settings except ANGLES, returns
                          the model that consists of the explanation.
                          If the method is ANGLES, returns a list of tuples
                          of the almost parallel rows and almost parallel
                          columns, and the model associated with the basis
                          matrix containing those almost parallel rows or
                          columns."""

    if model.IsMIP or model.IsQP or model.IsQCP:
        print("Ill Conditioning explainer only operates on LPs.")
        return None
    if _debug != OFF:
        if _debugger != OFF:
            import pdb

            pdb.set_trace()

    if expltype == "COLS":
        expltype = BYCOLS
    elif expltype == "ROWS":
        expltype = BYROWS
    else:
        print("Invalid specification for expltype:", str(expltype) + ": exiting.")
        return None

    if relobjtype == "QP":
        relobjtype = SOLVEQP
    elif relobjtype == "LP":
        relobjtype = SOLVELP
    elif relobjtype == "MIP":
        print("MIP subproblem currently unsupported; exiting. ")
        print(
            "Consider setting method to LASSO or submatrix parameter",
            "to False to obtain smaller explanation.",
        )
        return None
    else:
        print("Invalid specification for relobjtype:", str(relobjtype) + "; exiting.")
        return None

    if method == "DEFAULT":
        method = DEFAULT
    elif method == "ANGLES":
        method = ANGLES
    elif method == "LASSO":
        method = LASSO
    elif method == "RLS":
        method = RLS
    else:
        print("Invalid specification for method:", str(method) + "; exiting.")
        return None

    modvars = model.getVars()
    modcons = model.getConstrs()

    if method == ANGLES:
        return angle_explain(model, 1)
    #
    #   Threshold for ill conditioning is ratio of current feasibility
    #   tolerance divided by the current machine precision.  Calculate
    #   via logs to avoid dividing small number into larger one.
    #
    macheps = np.finfo(float).eps
    feastol = model.getParamInfo("FeasibilityTol")[2]
    texp = math.log(feastol, 2) - math.log(macheps, 2)
    condthresh = math.pow(2, texp)

    splitfreevars = method == LASSO
    explmodel, splitvardict, RSinginfo, CSinginfo, Singflag = extract_basis(
        model, modvars, modcons, expltype, method, condthresh
    )
    if explmodel == None:  # Didn't find a basis for original model
        return None  # Nothing to explain
    resmodel = None
    kappa_stats(model, data, KappaExact)
    if _debug != OFF:
        explmodel.write("explmodel.lp")
    if Singflag:
        #
        # Detected an explanation based on row or column singletons
        # with a huge ratio among two entries.  Use a separate, simpler
        # routine to create the explanation.
        #
        resmodel, maxmult, minmult = build_singleton_resmodel(explmodel, expltype)
        illcond_report(resmodel, model.ModelName, expltype, maxmult, minmult)
        return resmodel
    #
    #   Minimize the violations of B'y = 0 constraints.  Do not relax
    #   the e'y == 1 constraint, and the y variables are free, so they
    #   have no bounds to relax.
    #
    exvars = explmodel.getVars()
    excons = explmodel.getConstrs()
    nbas = len(excons) - 1
    #
    #   The feasRelax model involves an ill conditioned basis matrix at its
    #   core, so we take steps to help the LP solve deal with that.  Start
    #   with all the y variables superbasic, set NumericFocus to 3, and look at the
    #   row and column ratios of the basis matrix to decide if we should
    #   use geometric mean scaling.
    #
    #    for v in exvars:
    #        v.VBasis = GRB.SUPERBASIC
    explmodel.setParam("NumericFocus", 3)
    explmodel.setParam("Method", 0)
    explmodel.setParam("FeasibilityTol", 1e-9)
    explmodel.setParam("OptimalityTol", 1e-9)
    rowrat = 1.0
    for con in excons[0:nbas]:
        rowvals = []
        rowlhs = explmodel.getRow(con)
        if rowlhs.size() == 0:
            continue
        rowvals = [rowlhs.getCoeff(j) for j in range(rowlhs.size())]
        freqs = common.get_vector_frequencies(rowvals, 10)
        rat = freqs[len(freqs) - 1][0] - freqs[0][0]  # base 10 exponents
        if rat > rowrat:
            rowrat = rat
    if rowrat > 5:
        explmodel.setParam("ScaleFlag", 2)
    else:  # Row ratios don't need geometric mean scaling; check col ratios.
        colrat = 1.0
        for var in exvars:
            colvals = []
            col = explmodel.getCol(var)
            if col.size() == 0:
                continue
            colvals = [col.getCoeff(i) for i in range(col.size())]
            freqs = common.get_vector_frequencies(colvals, 10)
            rat = freqs[len(freqs) - 1][0] - freqs[0][0]  # base 10 exponents
            if rat > colrat:
                colrat = rat
        if colrat > 5:
            explmodel.setParam("ScaleFlag", 2)
    if prmfile != None:
        explmodel.read(prmfile)
    minrelax = method == LASSO or method == RLS
    explmodel.feasRelax(
        relobjtype, minrelax, None, None, None, excons[0:nbas], [1.0] * nbas
    )
    #
    #   Solve configuration completed.  Solve the model that will give us
    #   a certificate of ill conditioning.
    #
    if _debug != OFF:
        explmodel.write("explmodel_fr.mps")

    explmodel.optimize()
    status = explmodel.status
    if status in (
        GRB.INF_OR_UNBD,
        GRB.INFEASIBLE,
        GRB.UNBOUNDED,
        GRB.NUMERIC,
        GRB.SUBOPTIMAL,
    ):
        #
        # Under perfect precision, the feasrelax model we are solving
        # should never exit infeasible or unbounded.  In that case, or if
        # the status indicates numerical issues of some sort, we should
        # take even more precautions to deal with the ill conditioned
        # submatrix in the feasrelax problem.  We do this by starting with
        # the all slack basis.   The idea is we want to find the minimal
        # relaxation with as few structural variables in the basis as
        # possible.   We don't do this by default because it was found to
        # be too time consuming.
        #
        allvars = explmodel.getVars()
        structvars = allvars[0:nbas]
        slackvars = allvars[nbas:]
        for struct in structvars:
            struct.VBasis = GRB.SUPERBASIC
        for slack in slackvars:
            slack.VBasis = GRB.NONBASIC_LOWER
        cons = explmodel.getConstrs()
        for c in cons:
            c.CBasis = GRB.BASIC
        explmodel.optimize()

    #
    #   The y variables were created before calling feasRelax.  Extract
    #   them, as they are the certificate of ill conditioning.
    #
    #   TODO:  This subsequent section to build the resmodel that contains
    #   the explanation has grown large enough that it probably belongs in
    #   either one separate routine or two routines (one for building by
    #   rows and one for building by column.
    #
    yvaldict = {}
    ynamedict = {}
    yvars = None
    if expltype == BYROWS:
        yvars = explmodel.getVars()[0:nbas]
        resmodel = model.copy()
        #
        # Need original model basis statuses as well in order to compute
        # rhs values b - A_Nx_N for the constraints in the explanation
        # model.
        #
        resvars = resmodel.getVars()
        varsolinfodict = {}
        for ov in model.getVars():
            varsolinfodict[ov.VarName] = (ov.VBasis, ov.X)
        rcons = resmodel.getConstrs()
        rconsdict = {}
        delcons = []
        #
        # The order in which we created variables when extracting
        # the explainer problem typically will not match the order of the
        # constraints in the original model.   So we need to use a dictionary
        # to map the support of the y vector to the correct constraints
        # in the computed explanation.
        #
        combinedcnt = 0
        combinedrhs = 0.0
        combinedlhs = gp.LinExpr()  # y'A; y is inf. certificate
        for c in rcons:
            rconsdict[c.index] = c

        for yv in yvars:
            if splitfreevars:
                #
                # resmodel is in terms of original model.   Need to
                # provide y values and constraint names in terms of
                # the original model, not the explainer model in which
                # the free variables have been split.  Make use of the
                # splitvardict dictionary that connects each pair of
                # split free variables to do this.  We need to extract
                # the original model constraint name from the split
                # variable names that have the plus/minus GRB suffixes.
                #
                suffix_len = len("_GRBPlus")
                yval = yv.X - splitvardict[yv.index].X
                yname = yv.VarName[0:-suffix_len]  # original constr name.
            else:
                yval = yv.X
                yname = yv.VarName

            zerotol = smalltol
            thiscon = rconsdict[yv.index]
            rownorm = L1_rownorm(resmodel, thiscon)
            if rownorm > 0.0:  # Empty lhs is possible
                if smalltol == DEFSMALLTOL:
                    zerotol = min(smalltol, 1 / (10 * rownorm))
                    if zerotol < macheps:
                        zerotol = 0.0

            if abs(yval) <= zerotol:
                delcons.append(rconsdict[yv.index])  # To be filtered out.
            else:
                if _debug == VERBOSE:
                    print("Include constraint ", yname)

                thiscon = rconsdict[yv.index]
                explname = "(mult=" + str(yval) + ")" + thiscon.ConstrName
                yvaldict[explname] = abs(yval)
                thiscon.ConstrName = explname
                thislhs = resmodel.getRow(thiscon)
                thisrhs = thiscon.rhs
                #
                # Calculate rhs for the basis matrix rows.  Given by
                # Bx = b - A_Nx_N for the rows in the explanation.
                # Note that Gurobi allows variables to be nonbasic at
                # infinite bounds with dual infeasibilities in dual simplex
                # method, so need to handle that possible situation.
                #
                for j in range(thislhs.size()):
                    var = thislhs.getVar(j)
                    coeff = thislhs.getCoeff(j)
                    varstat = varsolinfodict[var.VarName][0]
                    if varstat == GRB.BASIC:
                        continue
                    if varstat == GRB.NONBASIC_LOWER:
                        if var.LB > -GRB.INFINITY:
                            thisrhs -= var.LB * coeff
                    elif varstat == GRB.NONBASIC_UPPER:
                        if var.UB < GRB.INFINITY:
                            thisrhs -= var.UB * coeff
                    elif varstat == GRB.SUPERBASIC:
                        thisrhs -= varsolinfodict[var.VarName][1] * coeff

                thiscon.rhs = thisrhs
                mult = yval
                combinedsense = GRB.EQUAL
                # Can be deleted if no issues arise.
                # if thiscon.sense == GRB.GREATER_EQUAL:
                #    mult         = -yval
                #    combinedcnt += 1
                # else:
                #    mult = yval
                #    if thiscon.sense == GRB.LESS_EQUAL:
                #        combinedcnt += 1
                # Inf. Certificate contribution to this constraint and rhs
                combinedlhs.add(thislhs, mult)
                combinedrhs += thisrhs * mult

        # if combinedcnt > 0:         can be deleted if no issues arise.
        #    combinedsense = GRB.LESS_EQUAL
        # else:
        #    combinedsense = GRB.EQUAL
        resmodel.addLConstr(combinedlhs, combinedsense, combinedrhs, COMBINEDROW)
        resmodel.remove(delcons)
        resmodel.update()
        #
        # Filter out the nonbasic variables of the original model that appear
        # in the constraints ih the explainer model.
        #
        delvars = []
        resvardict = {}
        resvars = resmodel.getVars()
        for v in resvars:
            resvardict[v.VarName] = v
        for v in model.getVars():
            if v.vBasis != GRB.BASIC:
                delvars.append(resvardict[v.VarName])
        resmodel.remove(delvars)
        resmodel.update()
        #
        # Print the combined value y'B to the screen, including all nonzero
        # values
        #
        combinedcon = (resmodel.getConstrs())[resmodel.numConstrs - 1]
        print(
            "Vector matrix product of certificate of ill conditioning" + " and basis:"
        )
        print(resmodel.getRow(combinedcon))
        refine_output(resmodel, yvaldict, expltype, submatrix)
    else:  # Column based explanation
        yvars = explmodel.getVars()[0:nbas]
        resmodel = explmodel.copy()
        resvars = resmodel.getVars()
        resvardict = {}
        delvars = []
        combinedcol = gp.Column()
        for v in resvars:
            resvardict[v.index] = v
        for yv in yvars:
            if splitfreevars:
                #
                # Unlike the row explanation, resmodel may include slack
                # variables that are not explicit variables of the original
                # model.   Need to provide y (i.e. column) values in terms of
                # the original model, not the explainer model in which
                # the free variables have been split.  Make use of the
                # splitvardict dictionary that connects each pair of
                # split free variables to do this.  We need to extract
                # the original model variable name from the split
                # variable names that have the plus/minus GRB suffixes.
                #
                minusvar = splitvardict[yv.index]
                yval = yv.X - minusvar.X
                ynamelen = len(yv.VarName) - len("_GRBPlus")
                yname = yv.VarName[0:ynamelen]  # original variable name
                delvars.append(resvardict[minusvar.index])
            else:
                yval = yv.X
                yname = yv.VarName

            thisvar = resvardict[yv.index]
            colnorm = L1_colnorm(resmodel, thisvar)
            #
            #           By default, with really large row values, we try to capture
            #           the significance of even really small row multipliers.
            #           But if caller specifies a non default tolerance, we don't
            #           overrule them regardless of the presence of really large
            #           coefficients that might be missed due to their choice of
            #           small multiplier tolerance.
            #
            if smalltol == DEFSMALLTOL:
                zerotol = min(smalltol, 1 / colnorm)
                if zerotol < macheps:
                    zerotol = 0.0
            else:
                zerotol = smalltol
            if abs(yval) < zerotol:
                delvars.append(resvardict[yv.index])  # To be filtered out.
            else:
                #
                # Don't include relaxation variables.
                #
                if _debug == VERBOSE:
                    print("Include variable ", yname)

                explname = ("(mult=" + str(yval) + ")") + yname
                thisvar.VarName = explname
                if method == LASSO:  # Bound changed to 0 for Lasso split vars
                    thisvar.LB = -math.inf
                yvaldict[explname] = abs(yval)
                thiscol = resmodel.getCol(thisvar)
                coefflist = []
                conlist = []
                for k in range(thiscol.size()):
                    coefflist.append(yval * thiscol.getCoeff(k))
                    conlist.append(thiscol.getConstr(k))
                combinedcol.addTerms(coefflist, conlist)

        #
        #       Remove the feasrelax slack variables that were introduced into the
        #       explainer model.  Then remove the variables from the explainer model
        #       that are not part of the explanation.
        #
        explmodvars = resmodel.getVars()
        start = nbas
        if splitfreevars:
            start = 2 * nbas
        end = len(explmodvars)
        resmodel.remove(explmodvars[start:end])
        # resmodel.remove(delvars)
        for dv in delvars:
            resmodel.remove(dv)

        resmodel.update()
        resmodel.addVar(name=COMBINEDCOL, column=combinedcol)
        rcons = resmodel.getConstrs()
        #
        # e'x = 1 is not in explanation
        # if using lasso approach, then feasrelax phase I constraint
        # also needs to be removed from the explanation
        #
        resmodel.remove(rcons[len(rcons) - 1])
        if splitfreevars:
            resmodel.remove(rcons[len(rcons) - 2])
        resmodel.update()
        refine_output(resmodel, yvaldict, expltype, submatrix)
        #
        # Done with column specific part of explanation.
        #

    illcond_report(
        resmodel,
        model.ModelName,
        expltype,
        max(yvaldict.values()),
        min(yvaldict.values()),
    )
    if method == ANGLES:
        return ([], [], None)  # Compatibility with method=ANGLES
    else:
        return resmodel


#
#   For a given basis matrix B from the model provided, create the basis
#   model:
#   B'y = 0         // or By = 0 if modeltype == BYCOLS
#   e'y = 1         // normalization of y != 0 constraint
#   y free
#
#   This is an infeasible model for any nonsingular basis matrix B.
#
def extract_basis(
    model, modvars, modcons, modeltype=BYROWS, method=DEFAULT, condthresh=1e10
):
    Singflag = False
    #
    #   Does the model have a factorized basis?  If not, need to solve it
    #   first.
    #
    need_basis = False
    try:
        test = model.Kappa
    except AttributeError as e:
        need_basis = True

    if need_basis:
        #
        # No factorized basis available. If basis statuses are available,
        # use the basis specified by the statuses, irrespective of optimality.
        #
        try:
            model.write("GRBjunk.bas")
            #
            # No factorization, but complete basis statuses available.
            # Compute the explanation on the basis in the statuses.
            # TODO: find a better way; this is ugly since it relies on
            # the ability to write to disk.
            #
            model.setParam("IterationLimit", 0)
            model.setParam("Method", 0)
            os.remove("./GRBjunk.bas")
        except gp.GurobiError as e:
            tmp = 0  # No statuses or factorization; solve from scratch

        model.optimize()
        try:  # Confirm basis available after solve.
            tmp2 = modvars[0].VBasis
            tmp1 = modvars[0].X
        except AttributeError as e:  # no basis; no explanation to return
            return None, None, None, None, None

    m = model.numConstrs
    n = model.numVars
    #
    #   Extract the basic structural variables from the model into the model
    #   containing the basis matrix of interest.  When explaining by rows,
    #   each column of the basis matrix corresponds to a constraint in the
    #   explainer model, and each row of the basis matrix corresponds to a
    #   variable in the explainer model.
    #   Refinement:  Consider the row and column singletons:
    #   B  = D1   0   0            block row 1 (row singletons)
    #        A1  D2  A2            block row 2 (column singletons)
    #        A3   0  A4            block row 3 (everything else)
    #   Look for special cases where D1 or D2 are individually
    #   ill conditioned (i.e. their ratio of largest to smallest
    #   coefficients exceeds the ill conditioning threshold).
    #   In those cases, just return D1 or D2 as the basis model,
    #   resulting in a much easier computation.
    #
    explmodel = gp.Model("basismodel")
    RSinginfo = []  # list of tuples for row singletons
    CSinginfo = []  # list of tuples for column singletons

    build_explmodel(model, explmodel, modvars, modcons, RSinginfo, CSinginfo, modeltype)
    #
    #   Check for special cases where ill conditioning resides just in the
    #   the row or column singleton matrices D1 or D2 (i.e. max to min diagonal
    #   element exceeds the condition number threshold (~1e+10 for 64 bit doubles)
    #
    # TODO: consolidate as much code as possible in this if/elif block.
    if modeltype == BYROWS and len(RSinginfo) > 1:
        maxrat = 0
        maxcoeff = 0
        mincoeff = float("inf")
        maxvar = None
        minvar = None
        maxcon = None
        mincon = None
        for diag in RSinginfo:
            t = abs(diag[COEFF])
            if t > maxcoeff:
                maxcoeff = t
                maxvar = diag[VAR]
                maxcon = diag[CON]
            if t < mincoeff:
                mincoeff = t
                minvar = diag[VAR]
                mincon = diag[CON]
        assert t > 0  # debug
        if maxcoeff / mincoeff > condthresh:
            #
            # Row singleton matrix D1 is ill conditioned by itself.  Just use
            # the 2x2 matrix associated with the max and min diagonal elements
            # for the explanation
            #
            explmodel.dispose()
            explmodel = gp.Model("rowsingleton_basismodel")
            build_explmodel(
                model, explmodel, [minvar, maxvar], [mincon, maxcon], None, None, BYROWS
            )
            #
            #
            # The 2x2 matrix is of the form  a1  0 |  0...0
            #                                0  a2 |  0...0
            #                                --------------
            #                                  A1  |   A2
            # where a1 and a2 are the min or max diagonals.
            # The associated feasrelax model has numerous alterate optimal
            # solutions with minimum relaxation value 1.0.  We seek the one
            # where both rows of the matrix have positive multipliers,
            # so we fix the multipliers to the optimal solution with both
            # multipliers at nonzero values.
            #

            denom = mincoeff + maxcoeff
            minmult = mincoeff / denom
            maxmult = maxcoeff / denom
            explmodel.update()
            for con in explmodel.getConstrs():
                lhs = explmodel.getRow(con)
                coeff = lhs.getCoeff(0)
                var = lhs.getVar(0)
                if abs(coeff) < maxcoeff:
                    var.LB = maxmult
                    var.UB = maxmult
                else:
                    var.LB = minmult
                    var.UB = minmult
            Singflag = True
    elif modeltype == BYCOLS and len(CSinginfo) > 1:
        maxrat = 0
        maxcoeff = 0
        mincoeff = float("inf")
        maxvar = None
        minvar = None
        maxcon = None
        mincon = None
        for diag in CSinginfo:
            t = abs(diag[COEFF])
            if t > maxcoeff:
                maxcoeff = t
                maxvar = diag[VAR]
                maxcon = diag[CON]
            if t < mincoeff:
                mincoeff = t
                minvar = diag[VAR]
                mincon = diag[CON]
        assert t > 0  # debug
        if maxcoeff / mincoeff > condthresh:
            #
            # Column singleton matrix D2 is ill conditioned by itself.  Just use
            # the 2x2 matrix associated with the max and min diagonal elements
            # for the explanation
            #
            explmodel.dispose()
            explmodel = gp.Model("colsingleton_basismodel")
            build_explmodel(
                model, explmodel, [minvar, maxvar], [mincon, maxcon], None, None, BYCOLS
            )
            #
            # The 2x2 matrix is of the form  a1  0 | A2(1,.)
            #                                0  a2 | A2(2,.)
            #                                --------------
            #                                0   0  |   A3
            # where a1 and a2 are the min or max diagonals.
            # The associated feasrelax model has numerous alterate optimal
            # solutions with minimum relaxation value 1.0.  We seek the one
            # where both rows of the matrix have positive multipliers,
            # so we fix the multipliers to the optimal solution with both
            # multipliers at nonzero values.
            #
            denom = mincoeff + maxcoeff
            minmult = mincoeff / denom
            maxmult = maxcoeff / denom
            explmodel.update()
            for var in explmodel.getVars():
                thiscol = explmodel.getCol(var)
                coeff = thiscol.getCoeff(0)
                if abs(coeff) < maxcoeff:
                    var.LB = maxmult
                    var.UB = maxmult
                else:
                    var.LB = minmult
                    var.UB = minmult

            Singflag = True
    #
    #   B'y = 0 or By = 0 constraints are done.   All variables are now created, so
    #   add the e'y = 1 constraint and we are done.  This is the same,
    #   regardless of whether the explanation model is row or column based.
    #
    explmodel.update()
    splitvardict = None
    if method == LASSO:
        plusvars = explmodel.getVars()
        splitvardict = split_mirroredvars(explmodel)
        minusvars = []
        for v in plusvars:
            minusvar = splitvardict[v.index]
            minusvars.append(minusvar)
        #            explmodel.addSOS(GRB.SOS_TYPE1, [v, minusvar])

        explmodel.addConstr(
            gp.quicksum(plusvars) - gp.quicksum(minusvars) == 1, name="GRBNormalize"
        )
        explmodel.setObjective(gp.quicksum(plusvars) + gp.quicksum(minusvars))
    else:
        explvars = explmodel.getVars()
        explmodel.addConstr(gp.quicksum(explvars) == 1, name="GRBNormalize")
        if method == RLS:
            numvars = len(explvars)
            sumsquares = gp.QuadExpr()
            sumsquares.addTerms(numvars * [1], explvars, explvars)
            explmodel.setObjective(sumsquares)

    explmodel.update()
    return explmodel, splitvardict, RSinginfo, CSinginfo, Singflag


#
#   Builds the explainer model.  modvars and modcons can be all variables
#   and constraints in the model, but can also be a subset.  Therefore,
#   this routine can be used to build the explainer model after some
#   variables and/or constraints have been filtered out from being
#   possible causes of ill conditioning.
#
def build_explmodel(
    model, explmodel, modvars, modcons, RSinginfo, CSinginfo, modeltype
):
    starttime = time.time()
    t1 = 0.0
    t2 = 0.0
    sumtime = 0.0
    if modeltype == BYROWS:
        #
        #   B'y = 0
        #   e'y = 1         // normalization of y != 0 constraint
        #   y free
        #
        explvardict = {}
        explcondict = {}
        #
        # Build a dictionary of variables for every constraint in the
        # original model.   Will remove any variables associated with
        # constraints not in modcons after the next loop.
        #

        if _debug:
            sumtime = 0.0
            t1 = time.time()

        conindexdict = {}
        for con in model.getConstrs():
            conindexdict[con.index] = con
            explvardict[con.index] = explmodel.addVar(
                lb=-float("inf"), name=con.ConstrName
            )
        if _debug:
            t2 = time.time()
            sumtime += t2 - t1
            print("inner for loop time = ", sumtime)

        varlist = []
        coeflist = []
        rowcounts = model.NumConstrs * [0]
        for var in modvars:  # structural basic variables
            if var.VBasis != GRB.BASIC:
                continue
            col = model.getCol(var)
            collen = col.size()
            if collen == 1:  # record column singleton
                if CSinginfo != None:
                    CSinginfo.append((col.getConstr(0), var, col.getCoeff(0)))
            for i in range(collen):
                coeff = col.getCoeff(i)
                con = col.getConstr(i)
                rowcounts[con.index] += 1
                varlist.append(explvardict[con.index])
                coeflist.append(coeff)

            lhs = gp.LinExpr(coeflist, varlist)
            #            explcondict[var.VarName] = explmodel.addConstr(lhs == 0, \
            #                                                           name=var.VarName)
            explmodel.addConstr(lhs == 0, name=var.VarName)
            coeflist.clear()
            varlist.clear()

        if RSinginfo != None:
            for ind in range(model.NumConstrs):
                if rowcounts[ind] == 1:
                    con = conindexdict[ind]
                    lhs = model.getRow(con)
                    var = lhs.getVar(0)
                    if var.VBasis == GRB.BASIC:
                        RSinginfo.append((con, var, lhs.getCoeff(0)))

        if CSinginfo != None:
            #
            # slack basic variables.  Ignore if processing row or column
            # singleton model (in which case CSinginfo is None).
            #
            for con in modcons:  # slack basic variables
                if con.CBasis != GRB.BASIC:  # TODO:replace next 6 lines with fn.
                    continue
                if con.Sense == GRB.GREATER_EQUAL:
                    coeff = -1.0
                else:
                    coeff = 1.0
                explmodel.addConstr(
                    coeff * explvardict[con.index] == 0,
                    name="GRB_slack_" + con.ConstrName,
                )
                #
                # Slacks are column singletons by definition; treat them as
                # part of column singleton diagonal matrix even if also a row
                # singleton.
                #
                CSinginfo.append((con, None, coeff))
        #
        # The previous two loops iterated over modvars from the original
        # model.  If modcons is a subset of the original model constraints,
        # the previous loop may have added variables to the explainer
        # model corresponding to constraints in the original model
        # that are not in the modcons array.  Therefore, remove any
        # variables in the explainer model corresponding to constraints
        # not in the modcons list passed into the routine.
        # passed into the routine.  The explainer model has constraint
        # names based on original model variable names, and variable names
        # based on original model constraint names.
        #
        explmodel.update()
        if len(modcons) < model.NumConstrs:
            modconlist = []
            for con in modcons:
                modconlist.append(con.index)
            delvars = []
            for explvar in explvardict.values():
                if explvar.index in modconlist:
                    continue
                delvars.append(explvar)
            explmodel.remove(delvars)
            explmodel.update()
    else:  # BYCOLS; assumes modeltype has been checked by caller
        #
        #   By = 0
        #   e'y = 1         // normalization of y != 0 constraint
        #   y free
        #
        explcondict = {}
        modvarlist = []
        conindexdict = {}
        for con in model.getConstrs():
            #
            # Constraint initialization for column based explainer model.
            # Build dictionary of constraints for every constraint
            # in the original mode. Will remove any constraints associated
            # with constraints not in modcons after the next loop.
            #
            conindexdict[con.index] = con
            explcondict[con.index] = explmodel.addLConstr(
                0, GRB.EQUAL, 0, name=con.ConstrName
            )

        #
        # Structural basic variables.  The column from the original model
        # needs to reference the constraint in the explainer model, not
        # the original model.
        #
        colcons = []
        colcoeffs = []
        rowcounts = model.NumConstrs * [0]
        for var in modvars:
            if var.VBasis != GRB.BASIC:
                continue
            col = model.getCol(var)
            collen = col.size()
            if collen == 1:  # record column singleton
                if CSinginfo != None:
                    CSinginfo.append((col.getConstr(0), var, col.getCoeff(0)))

            for k in range(collen):
                thiscon = col.getConstr(k)
                rowcounts[thiscon.index] += 1
                colcons.append(explcondict[thiscon.index])
                colcoeffs.append(col.getCoeff(k))
            explmodel.addVar(
                obj=0.0,
                lb=-float("inf"),
                name=var.VarName,
                column=gp.Column(colcoeffs, colcons),
            )
            colcons.clear()
            colcoeffs.clear()

        if RSinginfo != None:
            for ind in range(model.NumConstrs):
                if rowcounts[ind] == 1:
                    con = conindexdict[ind]
                    lhs = model.getRow(con)
                    var = lhs.getVar(0)
                    if var.VBasis == GRB.BASIC:
                        RSinginfo.append((con, var, lhs.getCoeff(0)))

        if CSinginfo != None:
            #
            # slack basic variables.  Ignore if processing row or column
            # singleton model (in which case CSinginfo is None).
            #
            for con in modcons:  # slack basic variables
                if con.CBasis != GRB.BASIC:  # TODO: replace next 6 lines with fn.
                    continue
                if con.Sense == GRB.GREATER_EQUAL:
                    coeff = -1.0
                else:
                    coeff = 1.0
                vname = "GRBslack_" + con.ConstrName
                col = gp.Column([coeff], [explcondict[con.index]])
                explmodel.addVar(obj=0.0, lb=-float("inf"), name=vname, column=col)
                #
                # Slacks are column singletons by definition; treat them as
                # part of column singleton diagonal matrix even if also a row
                # singleton.
                #
                CSinginfo.append((con, None, coeff))
        #
        # The previous two loops iterated over modvars from the original
        # model.  If modcons is a subset of the original model constraints,
        # the previous loop may have included constraints in the explainer
        # model corresponding to constraint in the original model not in
        # the modcons array.   Therefore, remove any constraints in the
        # explainer model that are not in the modcons list.  The explainer
        # model has variable and constraint names based on the original
        # model variable and constraint names, respectively.
        #
        explmodel.update()
        if len(modcons) < model.NumConstrs:
            modconlist = []
            for con in modcons:
                modconlist.append(con.index)
            delcons = []
            for explcon in explcondict.values():
                if explcon.index in modconlist:
                    continue
                delcons.append(explcon)
            explmodel.remove(delcons)
            explmodel.update()

    if _debug != OFF:
        endtime = time.time()
        print("Time build_explmodel = ", endtime - starttime)


#
#   Handle the special case where the explanation consists of just
#   two row singletons or column singletons with a large ratio.
#   No need to map back to the original model or run feasrelax
#   on the explainer model, which already has an optimal solution
#   provided since the variables are fixed.
#
def build_singleton_resmodel(explmodel, expltype):
    vars = explmodel.getVars()
    cons = explmodel.getConstrs()
    #
    # No need to optimize the explmodel since all its variables are fixed.
    #
    if vars[0].LB > vars[1].LB:
        maxmult = vars[0].LB
        minmult = vars[1].LB
        maxvar = vars[0]
        minvar = vars[1]
    else:
        maxmult = vars[1].LB
        minmult = vars[0].LB
        maxvar = vars[0]
        minvar = vars[1]

    if expltype == BYROWS:
        #
        # Transpose the explanation model into the results model.
        # For example, this explanation model  (from neos-799716.mps)
        #
        # cons(6,1): - 0.000576 c2931 = 0
        # beginFamily(0,3): - 1e+08 c14960 = 0
        # R2: c2931 + c14960 = 1
        # Bounds
        # c2931 = 1
        # c14960 = 5.76e-12
        #
        # would result in
        # (mult=0.9999999999942399)c2931: - 0.000576 cons(6,1) = 0
        # (mult=5.759999999966822e-12)c14960: - 1e+08 beginFamily(0,3) = 0
        #
        resmodel = gp.Model()
        maxconname = "(mult=" + str(maxmult) + ")" + maxvar.VarName
        col = explmodel.getCol(maxvar)
        con0 = col.getConstr(0)
        con1 = col.getConstr(1)
        if con1.ConstrName == "GRBNormalize":
            maxvarname = con0.ConstrName
            maxcoeff = col.getCoeff(0)
        else:
            maxvarname = con1.ConstrName
            maxcoeff = col.getCoeff(1)
        thisvar = resmodel.addVar(name=maxvarname)
        resmodel.addConstr(maxcoeff * thisvar == 0, name=maxconname)
        minconname = "(mult=" + str(minmult) + ")" + minvar.VarName
        col = explmodel.getCol(minvar)
        con0 = col.getConstr(0)
        con1 = col.getConstr(1)
        if con1.ConstrName == "GRBNormalize":
            minvarname = con0.ConstrName
            mincoeff = col.getCoeff(0)
        else:
            minvarname = con1.ConstrName
            mincoeff = col.getCoeff(1)
        thisvar = resmodel.addVar(name=minvarname)
        resmodel.addConstr(mincoeff * thisvar == 0, name=minconname)
    else:
        #
        # Column based explanation is easier, as no tranposition of the
        # model is needed.  Just modeify a copy of the explmodel.
        # For example, this explanation model will simply have its
        # names modified.
        #
        # c1311: 1e+11 x3259 = 0
        # c1312: x3260 = 0
        # GRBNormalize: x3260 + x3259 = 1
        # Bounds
        # x3260 = 1
        # x3259 = 1e-11
        #
        # would result in
        # At this point we no longer need the explmodel, so just modify
        # it in place.
        #
        maxvar.VarName = "(mult=" + str(maxmult) + ")" + maxvar.VarName
        minvar.VarName = "(mult=" + str(minmult) + ")" + maxvar.VarName
        cons = explmodel.getConstrs()
        for c in cons:
            if c.ConstrName == "GRBNormalize":
                explmodel.remove(c)
        explmodel.update()
        resmodel = explmodel
    resmodel.ModelName = explmodel.ModelName
    return resmodel, maxmult, minmult


def kappa_stats(model, data, KappaExact):
    kappa = model.Kappa
    kappaexact = -1
    if KappaExact == 1 and (model.numConstrs < 1000000 and model.numVars < 1000000):
        kappaexact = model.KappaExact
    print("--------------------------------------------------------")
    print("Condition number stats for basis for model ", model.ModelName, ":")
    print("Estimated condition number: ", kappa)
    if kappaexact == -1:
        print("Exact condition number not computed.")
    else:
        print("Exact condition number: ", kappaexact)
    print("--------------------------------------------------------")

    if data != None:
        data["Kappa"] = kappa
        data["KappaExact"] = kappaexact


#
#   Write out the final explanation, print summary info.
#
def illcond_report(resmodel, modelname, expltype, maxabsmult, minabsmult):
    resmodel.setObjective(0)
    if modelname == "":
        modelname = "model"
    if expltype == BYROWS:
        filename = modelname + "_kappaexplain.lp"
    else:
        filename = modelname + "_kappaexplain.mps"
    resmodel.write(filename)
    #
    #   Final info
    #
    print("--------------------------------------------------------")
    print("Ill conditioning explanation written to file ", filename)
    print("Maximum absolute multiplier value: ", maxabsmult)
    print("Minimum absolute multiplier value: ", minabsmult)
    print("--------------------------------------------------------")


#
#   Split free variables into difference of two nonnegative variables.
#   Also works on other variables whose lower bound is the opposite of
#   its upper bound.  varstosplit identifies the candidate list of variables
#   to split.  If not set or None, will treat all variables as candidates.
#   Only those that meet the "mirrored variable" criteria will
#   be split.   Most common use case will be to split free variables.
#
#   Alters the model, so use a copy if you want to preserve the original
#   model.
#
#   Tests:  1) Test on a model with only QCs.
#           2) Test on a model with linear obj terms but no linear constraints.
#
def split_mirroredvars(model, varstosplit=None):
    model.update()
    if varstosplit == None:
        varstosplit = model.getVars()

    newvarlist = []  # This var and dict is used to handle
    newvardict = {}  # quadratic objective and constraints
    for v in varstosplit:
        if v.UB != -v.LB:
            continue
        if v.UB == 0.0 and v.LB == 0.0:  # No need to split vars
            continue  # fixed at zero
        #
        #  We have a candidate variable to split.
        #
        bnd = v.UB
        varname = v.VarName
        chgvarname = v.VarName + "_GRBPlus"
        v.VarName = chgvarname
        v.LB = 0
        col = model.getCol(v)
        splitcol = gp.Column()
        coeflist = []
        conlist = []
        for k in range(col.size()):
            coeflist.append(-col.getCoeff(k))
            conlist.append(col.getConstr(k))
        if col.size() > 0:
            splitcol.addTerms(coeflist, conlist)
        newvar = model.addVar(
            lb=0.0,
            ub=bnd,
            obj=-v.obj,
            vtype=v.Vtype,
            name=varname + "_GRBMinus",
            column=splitcol,
        )
        newvarlist.append(newvar)
        newvardict[v.index] = newvar

    #
    # Linear constraints completed; now update any QCs that contain
    # mirrored variables that need to be split.
    #
    for qcon in model.getQConstrs():
        quadexpr = model.getQCRow(qcon)
        split_quadexpr(quadexpr, newvardict)
        #
        # Quadratic expression is updated; need to create a new
        # QC with the update quadexpr and delete the old one.
        #
        model.addQConstr(quadexpr, qcon.QCSense, qcon.QCRHS, qcon.QCName)
        model.remove(qcon)

    #
    # Quadratic constraints completed; now the quadratic objective if it
    # contains any mirrored variables that need to be split.
    #
    objexpr = model.getObjective()
    if isinstance(objexpr, gp.QuadExpr):
        split_quadexpr(objexpr, newvardict)
        model.setObjective(objexpr)
    model.update()
    return newvardict


#
#   Takes a quadratic expression and replaces mirrored variables with
#   the appropriate difference of two nonnegative variables.  All
#   mirrored variables are provided in newvardict, which maps the
#   original mirrored variable to its added variable.


def split_quadexpr(quadexpr, newvardict):
    for k in range(quadexpr.size()):
        xi = quadexpr.getVar1(k)
        xj = quadexpr.getVar2(k)
        qcoef = quadexpr.getCoeff(k)
        if xi.index in newvardict:  # xi is split into xi(+) - xi(-)
            # xi(-) * xj term
            quadexpr.addTerms(-qcoef, xj, newvardict[xi.index])
            if xj.index in newvardict:  # both xi and xj split
                # xi * xj(-) term
                quadexpr.addTerms(-qcoef, xi, newvardict[xj.index])
                # xi(-) * xj(-) term
                quadexpr.addTerms(qcoef, newvardict[xi.index], newvardict[xj.index])
        elif xj.index in newvardict:  # xj is split, but xi is not.
            # xi * xj(-) term
            quadexpr.addTerms(-qcoef, xi, newvardict[xj.index])


#
#   Look for easy explanations involving almost parallel pairs of
#   matrix rows or columns.   Any pair that is almost parallel can
#   be identified from having a very small angle, i.e., if the
#   difference between the inner product of two vectors and the product
#   if their two norms is within a small tolerance.  Or, if the negative
#   of the inner product and the two norms is within a tolerance, they
#   are almost parallel as well, i.e., the angle is close to 180 degrees.
#   The howmany parameter controls how many near parallel items to report.
#   Default is 1; integer values > 1 specify how many to report; integer
#   values <= 0 indicate report all.  TODO:  This should be callable
#   by the user, so it needs help info.
#
def angle_explain(model, howmany=1, partol=1e-6):
    #
    #   Help function info
    #
    """Specialized ill conditioning explainer for finding pairs of near
    parallel rows and near parallel columns of the basis matrix.
    Any basis statuses, in the model will be used, computing the
    factorization if needed.  If no statuses are available, solve the LP
    to optimality (or whatever the final solve status is), and use
    that basis.

    Arguments:
    model      (required) The LP model whose basis will be examined.
                          Basis can be from completed solve or statuses.
                          LP will be optimized if no basis is present.
    howmany    (optional) Number of pairs of near parallel rows and columns
                          to search for.  Default is 1.  Specify a positive
                          integer for a particular number of pairs.
                          Specify an integer <= 0 to request all pairs.
                          Routine searches rows first, then columns.
    partol     (optional) Tolerance below which two row or column vectors
                          are considered almost parallel.  Used as a
                          relative tolerance when comparing the inner
                          product of two vectors with the product of their
                          L1 norms.
    Returns:              returns a list of tuples of the almost parallel
                          rows and almost parallel columns, and the model
                          associated with the basis matrix containing those
                          almost parallel rows or columns."""

    if _debug != OFF:
        if _debugger != OFF:
            import pdb

            pdb.set_trace()
    modcons = model.getConstrs()
    modvars = model.getVars()
    explmodel, splitvardict, junk1, junk2, junk3 = extract_basis(
        model, modvars, modcons, None, False
    )
    if explmodel == None:  # No basis for original model found.
        return None, None, None  # Nothing to explain
    modcons = explmodel.getConstrs()
    modvars = explmodel.getVars()
    explmodel.remove(modcons.pop(-1))  # Just want basis matrix
    explmodel.update()
    #
    # Calculate mod value for hashing function, which depends on
    # max possible number of bits for an integer (typically 64).
    # Also set up some dictionaries for hashing calculations.
    #
    modval = 1 + math.ceil(math.log(sys.maxsize, 2))
    rowhashdict = {}
    for con in modcons:
        rowhashdict[con.ConstrName] = 0
    colhashdict = {}
    for var in modvars:
        colhashdict[var.VarName] = 0

    #
    # Look for almost parallel rows.   Hash the rows as follows.
    # For each row define a 64 bit int h[i] such that if variable
    # xj appears in the row with a coefficient larger than epsilon
    # you set bit j % 64 in h_i. Then, for a pair (i1,i2) of rows,
    # if h[i1] != h[i2], the two rows cannot be almost parallel.
    # First, create the hash values and put the constraints with the
    # same hash value in the same bucket.   Don't hash small nonzeros
    # below partol; that way we pick up the case where one row is
    # (1 1) and another intersects the same two nonzeros but is
    # (1 1 eps).   We want to flag those as almost parallel, so we
    # want them to have the same hash value.  We do this in a normalized
    # fashion, so that we will get the same conclusion for (1 1), (1, 1 eps)
    # and (1000 1000), (1000, 1000, 1000*eps).  Similarly, we don't
    # consider rows of the form (1e-5, 1e-5, 0) and (1e-5, 1e-5, 1e-7)
    # almost parallel since they are significantly different in a relative
    # sense.
    #
    bucketlists = [[] for i in range(modval)]
    for con in modcons:
        rowlhs = explmodel.getRow(con)
        normvec = []
        for j in range(rowlhs.size()):
            normvec.append(rowlhs.getCoeff(j))
        rownorm = L1_norm(normvec)

        for j in range(rowlhs.size()):
            var = rowlhs.getVar(j)
            coef = rowlhs.getCoeff(j)
            if abs(coef) > partol * rownorm:
                varbit = var.index % (modval - 1)
                rowhashdict[con.ConstrName] |= 1 << varbit
        bucketlists[rowhashdict[con.ConstrName] % (modval - 1)].append(con)
    #
    # Check the matrix rows in the same bucket to see if they are
    # almost parallel.
    #
    almostparcount = 0
    almostparrowlist = []
    quittingtime = False
    starttime = time.time()
    for bucket in bucketlists:
        for k in range(len(bucket)):
            lhs1 = explmodel.getRow(bucket[k])
            size1 = lhs1.size()
            norm1 = 0.0
            row1array = explmodel.numVars * [0.0]
            #
            # Treat first row as a dense vector, second row as a sparse one.
            #
            for i in range(size1):
                ind1 = lhs1.getVar(i).index
                coef1 = lhs1.getCoeff(i)
                row1array[ind1] = coef1
                norm1 += coef1**2
            for j in range((k + 1), len(bucket)):
                lhs2 = explmodel.getRow(bucket[j])
                size2 = lhs2.size()
                if size1 != size2:
                    continue
                #
                # Simple length based filtering failed.  Compute inner products
                # and norms.  Need to handle the case where two rows intersect
                # the same variables, but in a different order.
                #
                #
                # Compute the angle.  TODO: Make this calculation more
                # precise.  Consider Kahan summation, and avoid adding
                # big numbers and much smaller ones.
                #
                dotprod = 0.0
                norm2 = 0.0
                skip = False
                for i in range(size2):
                    ind2 = lhs2.getVar(i).index
                    if row1array[ind2] != 0.0:
                        coef2 = lhs2.getCoeff(i)
                        dotprod += row1array[ind2] * coef2
                        norm2 += coef2**2
                    else:  # Intersecting variables differ; nothing to see here
                        skip = True
                        break

                if skip:  # Early exit due to index or coeff mismatch
                    continue
                norm1 = math.sqrt(norm1)
                norm2 = math.sqrt(norm2)
                #
                # Normalize the tolerance in the test; want the same
                # result for vectors u and v and 1000u and 1000v
                #
                if abs(abs(dotprod) - norm1 * norm2) > partol * norm1:
                    continue
                #
                # Found two almost parallel rows.
                #
                almostparcount += 1
                almostparrowlist.append((bucket[k], bucket[j]))
                if howmany <= 0 or almostparcount < howmany:
                    continue
                quittingtime = True
                break  # Exit for j loop
            if quittingtime:
                break  # Exit for k loop
        if quittingtime:
            break  # Exit for bucket loop

    if _debug != OFF:
        endtime = time.time()
        print("Time for main row based loop = ", endtime - starttime)
    del row1array
    #
    # Now look for almost parallel columns.   Hash the rows in a smaller
    # manner, except this time the bit values are based on row indices
    # rather than column indices.
    #
    bucketlists = [[] for i in range(modval)]
    for var in modvars:
        col = explmodel.getCol(var)
        normvec = []
        for i in range(col.size()):
            normvec.append(col.getCoeff(i))
        colnorm = L1_norm(normvec)
        for i in range(col.size()):
            con = col.getConstr(i)
            coef = col.getCoeff(i)
            if abs(coef) > partol * colnorm:
                varbit = con.index % (modval - 1)
                colhashdict[var.VarName] |= 1 << varbit
        bucketlists[colhashdict[var.VarName] % (modval - 1)].append(var)
    #
    # Check the matrix columns in the same bucket to see if they are
    # almost parallel.
    #
    almostparcollist = []
    quittingtime = False
    starttime = time.time()
    for bucket in bucketlists:
        for k in range(len(bucket)):
            col1 = explmodel.getCol(bucket[k])
            col1array = explmodel.numVars * [0.0]
            norm1 = 0.0
            size1 = col1.size()
            for i in range(size1):  # Treat first column as dense vector.
                ind1 = col1.getConstr(i).index
                coef1 = col1.getCoeff(i)
                col1array[ind1] = coef1
                norm1 += coef1**2
            for j in range((k + 1), len(bucket)):
                col2 = explmodel.getCol(bucket[j])
                size2 = col2.size()
                if size1 != size2:
                    continue
                #
                # Simple length based filtering failed.  Compute inner products
                # and norms.  Need to handle the case where two columns
                # intersect the same rows, but in a different order.
                #
                # Compute the angle.  TODO: Make this calculation more
                # precise.  Consider Kahan summation, and avoid adding
                # big numbers and much smaller ones.
                #
                dotprod = 0.0
                norm2 = 0.0
                skip = False
                for i in range(size2):
                    ind2 = col2.getConstr(i).index
                    if col1array[ind2] != 0.0:
                        coef2 = col2.getCoeff(i)
                        dotprod += col1array[ind2] * coef2
                        norm2 += coef2**2
                    else:
                        #
                        # Intersecting constraints differ; move along, nothing
                        # to see here.
                        #
                        skip = True
                        break

                if skip:  # Early exit due to index or coeff mismatch
                    continue
                norm1 = math.sqrt(norm1)
                norm2 = math.sqrt(norm2)
                #
                # Normalize the tolerance in the test; want the same
                # result for vectors u and v and 1000u and 1000v
                #
                if abs(abs(dotprod) - norm1 * norm2) > partol * norm1:
                    continue
                #
                # Found two almost parallel rows.
                #
                almostparcount += 1
                almostparcollist.append((bucket[k], bucket[j]))
                if howmany <= 0 or almostparcount < howmany:
                    continue
                quittingtime = True
                break  # Exit for j loop
            if quittingtime:
                break  # Exit for k loop
        if quittingtime:
            break  # Exit for bucket loop

    if _debug != OFF:
        endtime = time.time()
        print("Time for main column based loop = ", endtime - starttime)

    return almostparrowlist, almostparcollist, explmodel


#
#   Utility routine to facilitate examination of explainer output.
#   Provides a bitmap of the explanation matrix.  Requires matplotlib
#   to be on the machine and to be able to display output.
#
def matrix_bitmap(model):
    #
    #   Help function info
    #
    """Utility routine to facilitate examination of explainer output by
    printing a bitmap of the nonzero structure.   Can also be used to
    obtain bitmaps for matrices of arbitrary models rather than the
    explainer model.


    Arguments:
    model      (required) The model whose constraint matrix will generate
                          the bitmap.  Note that the routine may remove
                          empty rows and columns from the constraint matrix
                          of the model, so make a copy of the model before
                          calling the routine if necessary.
    Returns:              Nothing.  The bitmap appears on the screen and
                          can be saved to a file."""
    try:
        import matplotlib.pyplot as plt
    except ImportError as e:
        print("Need matplotlib.pyplot installed to display bitmap.")
        return
    #
    #
    #
    #   We only want the bitmap for variables that intersect at least one
    #   constraint.
    #
    varstodelete = []
    for var in model.getVars():
        col = model.getCol(var)
        if col.size() == 0:
            varstodelete.append(var)
    model.remove(varstodelete)
    model.update()
    #
    #   Similarly, we only want constraints that intersect at least one
    #   variable.
    #
    constodelete = []
    for con in model.getConstrs():
        row = model.getRow(con)
        if row.size() == 0:
            constodelete.append(con)
    model.remove(constodelete)
    model.update()

    if model.ModelName == "":
        modelname = "model"
    else:
        modelname = model.ModelName
    if _debug != OFF:
        model.printStats()
    #
    #   Now do the plot.
    #
    A = model.getA()
    plt.spy(A, markersize=1)
    plt.title(modelname)
    plt.show()


#
#   Utility function that takes a list of decimal values and
#   uses the Fraction package to provide a rational approximation.
#
from fractions import Fraction


def converttofractions(vals):
    #
    #   Help function info
    #
    """Utility routine to convert decimal values to fractions.

    Arguments:
    vals       (required) Array of decimal values to convert
    Returns:              Nothing; prints results to screen."""
    for v in vals:
        if not isinstance(v, float):
            print("Value ", v, " is not a float.  Cannot convert.")
            continue
        frac = Fraction(v).limit_denominator()
        print(f"The approx fraction of {v} is {frac}.")


#
#   Improve the readability of the output
#   1) List the constraints in the explanation sorted by largest
#      absolute row multiplier first.
#   2) If model has row or column singletons, include only selected
#      submatrices in the explanations (details in comments below).
#
def refine_output(model, absmultdict, expltype, submatrix):
    #
    #   model is the explainer model, not the original one.
    #   This routine modifies the model, with the constraints sorted in
    #   descending order of the values in the absmultdict dictionary.
    #
    #    import pdb; pdb.set_trace()
    quit = False
    if expltype == BYROWS:
        removeemptycolumns(model)
        condict = {}
        #
        # Skip the combined constraint (the last one); it has no multiplier
        #
        cons = model.getConstrs()[0 : model.numConstrs - 1]
        concnt = 0
        for c in cons:
            condict[c.ConstrName] = c
            concnt += 1
        #
        # Sort the dictionary by absolute row multiplier value from
        # highest to lowest.  Dictionaries cannot be sorted directly,
        # so we use the sorted function to obtain a list of tuples
        # sorted by the dictionary value, then transform that list of
        # tuples back into a dictionary.  Details are at
        # https://www.freecodecamp.org/news/sort-dictionary-by-value-in-python/
        #
        sorteddict = dict(sorted(absmultdict.items(), key=lambda x: x[1], reverse=True))
        #
        # Add the constraints in sorted order to the end of the model, then
        # remove their duplicates in unsorted order that are the first concnt
        # constraints
        #
        for conname in sorteddict.keys():
            contoadd = condict[conname]
            lhs = model.getRow(contoadd)
            model.addLConstr(
                lhs, contoadd.sense, contoadd.rhs, name=contoadd.ConstrName
            )
        model.remove(cons[0:concnt])
        model.update()
    elif expltype == BYCOLS:
        removeemptyrows(model)
        vardict = {}
        #
        # Skip the combined column (the last one); it has no multiplier
        #
        vars = model.getVars()[0 : model.numVars - 1]
        varcnt = len(vars)
        for v in vars:
            vardict[v.VarName] = v
        #
        # Sort the dictionary by absolute col multiplier value from
        # highest to lowest.   Analogous to row multiplier sorting described
        # in detail in the comment in the BYROWS code above
        #
        sorteddict = dict(sorted(absmultdict.items(), key=lambda x: x[1], reverse=True))
        #
        # Add the variables in sorted order to the end of the model, then
        # remove their duplicates in unsorted order that are the first varcnt
        # variables
        #
        for varname in sorteddict.keys():
            vartoadd = vardict[varname]
            model.addVar(
                lb=vartoadd.lb,
                ub=vartoadd.ub,
                column=model.getCol(vartoadd),
                name=vartoadd.VarName,
            )
        model.remove(vars[0:varcnt])
        model.update()
    #
    # First refinement complete.  Model constraints or variables will
    # now be listed by largest absolute multiplier value first.
    #

    if submatrix:
        #
        # Additional reductions in explanation that may permit it to
        # be an ill conditioned submatrix of the original basis matrix.
        # Details depend on whether explanation is row or column based.
        #
        #   B  = D1   0   0            block row 1 (row singletons)
        #        A1  D2  A2            block row 2 (column singletons)
        #        A3   0  A4            block row 3 (everything else)
        #
        # Note that if we arrive here, preprocessing has already determined
        # that D1 and D2 individually are not ill conditioned.
        #
        # First, perform tasks common to both row and column based refinements.
        # Build lists of variables and constraints in the various
        # blocks.  Original singleton data structures RSinginfo and CSinginfo
        # were computed during basis extraction, so we need to remove the
        # associated constraints and variables that are not part of the
        # actual explanation.  Easier to just rebuild these data structures
        # than use the original ones.
        #
        cons = model.getConstrs()
        vars = model.getVars()
        D1condict = {}
        D1vardict = {}
        D2condict = {}
        D2vardict = {}
        D1coefdict = {}
        D2coefdict = {}
        for c in cons:
            lhs = model.getRow(c)
            if lhs.size() != 1:
                continue
            #
            # Row singleton found
            #
            var = lhs.getVar(0)
            if var.VarName.startswith("GRB_slack"):
                #
                # Treat row singleton slacks as column singletons
                # rather than row singletons.
                #
                D2condict[c.ConstrName] = c
                D2vardict[var.VarName] = None
                D2coefdict[var.VarName] = 1.0  # sign doesn't matter
            else:  # Row singleton, part of D1
                D1condict[c.ConstrName] = c
                D1vardict[var.VarName] = var
                D1coefdict[c.ConstrName] = lhs.getCoeff(0)

        for var in vars:
            if var.VarName.startswith("GRB_slack"):
                continue  # Slack column singletons already processed.
            col = model.getCol(var)
            if col.size() != 1:
                continue
            #
            # Record column singleton info
            #
            con = col.getConstr(0)
            D2condict[con.ConstrName] = con
            D2condict[var.VarName] = var
            D2coefdict[var.VarName] = col.getCoeff(0)

        if expltype == BYROWS:
            #
            #   B  = D1   0   0            block row 1 (row singletons)
            #        A1  D2  A2            block row 2 (column singletons)
            #        A3   0  A4            block row 3 (everything else)
            #
            # Reduction #1:  Let w1, w2 and w3 be the row multipliers from the
            # certificate of ill conditioning associated with blocks 1, 2 and
            # 3 respectively.   If w3'A4 shows that A4 is ill conditioned, then
            # we know that B is ill conditioned by setting w2 = 0,
            # w1 = -w3'D1^(-1)A3.   So just output the rows and columns in A4 by
            # removing the variables and constraints associated with blocks 1
            # and 2, and use the remaing  submatrix A4 as the explanation.
            #
            combocon = None
            for c in cons:
                if c.ConstrName == COMBINEDROW:
                    combocon = c
                    break
            combolhs = model.getRow(combocon)
            A4combovars = []
            A4combocoeffs = []
            for j in range(combolhs.size()):
                var = combolhs.getVar(j)
                if var.VarName in D1vardict or var.VarName in D2vardict:
                    continue
                #
                # This term of the combined constraint is a column of A4
                #
                A4combovars.append(var)
                A4combocoeffs.append(abs(combolhs.getCoeff(j)))  # contains w3'A4
            #
            # We have w3'A4.  We need w3 so we can assess near singularity
            # of A4. But make sure A4 has positive dimension before doing so.
            #
            if len(D1condict) + len(D2condict) < model.NumConstrs:
                w3vals = []
                for c in cons:
                    if c.ConstrName in D1condict or c.ConstrName in D2condict:
                        continue
                    if c.ConstrName.startswith(COMBINEDROW):
                        continue
                    w3vals.append(absmultdict[c.ConstrName])
                w3norm = L1_norm(w3vals)
                A4combonorm = L1_norm(A4combocoeffs)
                if A4combonorm / w3norm < 0.01:
                    #
                    # A4 ill conditioned.  Remove the constraints in block rows
                    # 1 and 2, and the variables in D1 and D2 from the explainer
                    # model.
                    #
                    delcons = list(D1condict.values())
                    delcons.append(list(D2condict.values()))
                    delvars = list(D1vardict.values())
                    delvars.append(list(D2vardict.values()))
                    model.remove(delcons)
                    model.update()
                    model.remove(delvars)
                    model.update()
                    #
                    # The remaining explanation model is just the variables
                    # and constraints in A4.   Don't look for additional
                    # refinements.
                    #
                    quit = True
        elif expltype == BYCOLS:
            #
            #   B  = D1   0   0            block row 1 (row singletons)
            #        A1  D2  A2            block row 2 (column singletons)
            #        A3   0  A4            block row 3 (everything else)
            #
            #
            # Reduction #1:  Let w1, w2 and w3 be the column multipliers
            # from the certificate of ill conditioning associated with
            # blocks 1, 2 and 3 respectively.   If A4w3 shows that A4
            # is ill conditioned, then we know that B is ill conditioned
            # by setting w1 = 0, w2 = -D2^(-1)A2. So just output the rows
            # and columns in A4 by removing the variables and constraints
            # associated with blocks 1 and 2, and use the remaining
            # submatrix A4 as the explanation.
            #
            combovar = None
            for v in vars:
                if v.VarName == COMBINEDCOL:
                    combovar = v
                    break
            combocol = model.getCol(combovar)
            A4combocons = []
            A4combocoeffs = []
            for i in range(combocol.size()):
                con = combocol.getConstr(i)
                if con.ConstrName in D1condict or con.ConstrName in D2condict:
                    continue
                #
                # This term of the dombined column is a row of A4
                #
                A4combocons.append(con)
                A4combocoeffs.append(abs(combocol.getCoeff(i)))
            #
            # We have A4w3.  We need w3 so we can assess near singularity of
            # A4.   But make sure  A4 has positive dimension before doing so.
            #
            if len(D1vardict) + len(D2vardict) < model.NumVars:
                w3vals = []
                for v in vars:
                    if v.VarName in D1vardict or v.VarName in D2vardict:
                        continue
                    if v.VarName.startswith(COMBINEDCOL):
                        continue
                    w3vals.append(absmultdict[v.VarName])
                w3norm = L1_norm(w3vals)
                A4combonorm = L1_norm(A4combocoeffs)
                if A4combonorm / w3norm < 0.01:
                    #
                    # A4 ill conditioned.  Remove the constraints and variables
                    # associated with blocks 1 and 2.
                    # TODO: If this stays indentical to the corresponding
                    # code in the BYROWS refinement, put this block into
                    # a single function.
                    #
                    delcons = list(D1condict.values())
                    delcons.append(list(D2condict.values()))
                    delvars = list(D1vardict.values())
                    delvars.append(list(D2vardict.values()))
                    model.remove(delcons)
                    model.update()
                    model.remove(delvars)
                    model.update()
                    #
                    # The remaining explanation model is just the variables
                    # and constraints in A4.   Don't look for additional
                    # refinements.
                    #
                    quit = True

    if not quit:
        #
        # Preprocessing checked whether D1 or D2 individually was ill
        # conditioned and took care of the explanation.   Here we check
        # if D1 and D2 combined are ill conditioned even though individually
        # they are well conditioned (e.g., D1 has all values around 1e-5
        # while D2 has all values around 1e+5).
        #
        quit = True


#
#   Improve the readability of the output.
#   1) List the variables in the explanation sorted by largest
#      absolute column multiplier first.
#
def refine_col_output(model, abscolmultdict):
    #
    #   model is the explainer model, not the original one.
    #   This routine modifies the model, with the variables sorted in
    #   descending order of the values in the abscolmultdict dictionary.
    #
    #    import pdb; pdb.set_trace()
    vardict = {}
    #
    # Skip the combined column (the last one); it has no multiplier
    #
    vars = model.getVars()[0 : model.numVars - 1]
    varcnt = len(vars)
    for v in vars:
        vardict[v.VarName] = v
    #
    # Sort the dictionary by absolute col multiplier value from
    # highest to lowest.  Dictionaries cannot be sorted directly,
    # so we use the sorted function to obtain a list of tuples
    # sorted by the dictionary value, then transform that list of
    # tuples back into a dictionary.  Details are at
    # https://www.freecodecamp.org/news/sort-dictionary-by-value-in-python/
    #
    sorteddict = dict(sorted(abscolmultdict.items(), key=lambda x: x[1], reverse=True))
    #
    # Add the variables in sorted order to the end of the model, then
    # remove their duplicates in unsorted order that are the first varcnt
    # variables
    #
    for varname in sorteddict.keys():
        vartoadd = vardict[varname]
        model.addVar(
            lb=vartoadd.lb,
            ub=vartoadd.ub,
            column=model.getCol(vartoadd),
            name=vartoadd.VarName,
        )
    model.remove(vars[0:varcnt])
    model.update()
    #
    # Model variables will now be listed by largest absolute multiplier
    # value first.
    #


#
#   Delete linear constraints that have no nonzero coefficients
#   Changes the model, so pass in a copy of you want to
#   preserve the original model.
#
def removeemptyrows(model):
    cons = model.getConstrs()
    delcons = []
    for c in cons:
        row = model.getRow(c)
        if row.size() == 0:
            delcons.append(c)
    model.remove(delcons)
    model.update()


#
#   Delete variables that have no nonzero coefficients in the
#   associated column of the linear constraint matrix.
#   Changes the model, so pass in a copy of you want to
#   preserve the original model.
#
def removeemptycolumns(model):
    vars = model.getVars()
    delvars = []
    for v in vars:
        col = model.getCol(v)
        if col.size() == 0:
            delvars.append(v)
    model.remove(delvars)
    model.update()


#
#   L1 norm of a specific constraint in a model.
#
def L1_rownorm(model, con):
    lhs = model.getRow(con)
    normvec = []
    for j in range(lhs.size()):
        normvec.append(lhs.getCoeff(j))
    result = L1_norm(normvec)
    return result


#
#   L1 norm of a specific variable in a model.
#
def L1_colnorm(model, var):
    col = model.getCol(var)
    normvec = []
    for j in range(col.size()):
        normvec.append(col.getCoeff(j))
    result = L1_norm(normvec)
    return result


#
#   L1 norm of an arbitrary vector.  Don't need super accurate
#   values here, so don't worry about Kahan summation or other
#   tactics for precision
#
def L1_norm(vec):
    sum = 0.0
    for v in vec:
        sum += abs(v)
    return sum


#
#   TODO: include Skeel condition number calculation
#


#
#   Debug routine
#
def bucket_info(bucketlists):
    histogram = []
    for bucket in bucketlists:
        histogram.append(len(bucket))
    return histogram



import gurobipy as gp

import gurobi_modelanalyzer.common as common


def process_variable_types(m, data, objective_variables, constraint_variables):
    print("Processing model variables...")

    # Count variable type
    variable_types = {
        gp.GRB.INTEGER: "Gen",
        gp.GRB.BINARY: "Bin",
        gp.GRB.CONTINUOUS: "Con",
        gp.GRB.SEMICONT: "SemiCont",
        gp.GRB.SEMIINT: "SemiInt",
    }

    variable_count = {VType: 0 for VType in variable_types.keys()}

    variable_boundtypes = ["Free", "LBOnly", "UBOnly", "Bnd", "Fixed"]
    variable_boundcount = {BType: 0 for BType in variable_boundtypes}

    redundant_vars = []

    for var in m.getVars():
        variable_count[common.get_vtype(var)] += 1

        if var.Obj == 0:
            if m.NumPWLObjVars > 0:
                pwl = m.getPWLObj(var)
                if len(pwl) == 0:
                    column = m.getCol(var)
                    if column.size() == 0:
                        redundant_vars.append(var)

        if var.LB == -gp.GRB.INFINITY:
            if var.UB == gp.GRB.INFINITY:
                variable_boundcount["Free"] += 1
            else:
                variable_boundcount["UBOnly"] += 1

        else:
            if var.UB == gp.GRB.INFINITY:
                variable_boundcount["LBOnly"] += 1
            else:
                if var.LB == var.UB:
                    variable_boundcount["Fixed"] += 1
                else:
                    variable_boundcount["Bnd"] += 1

    for vtype, count in variable_count.items():
        data["num%sVars" % variable_types[vtype]] = count

    for btype, count in variable_boundcount.items():
        data["num%sVars" % btype] = count

    # Check for redundant variables (not in objective, not in any constraint)
    redundant_vars = list(
        set(m.getVars()).difference(objective_variables, constraint_variables)
    )
    data["numRedundantVars"] = len(redundant_vars)
    data["redundantVars"] = [x.VarName for x in redundant_vars[:10]]



import unittest
import pathlib
import os

import gurobipy as gp
from gurobi_modelanalyzer import kappa_explain

here = pathlib.Path(__file__).parent
cwd = pathlib.Path(os.getcwd())


class TestExplainer(unittest.TestCase):
    """Super simple tests: did we write output files?"""

    def setUp(self):
        self.clean()
        self.env = gp.Env()
        self.mip_model = gp.read(str(here / "dataset" / "p0033.lp"), env=self.env)
        self.model = self.mip_model.relax()
        self.model.ModelName = "testmodel"

    def tearDown(self):
        self.model.close()
        self.mip_model.close()
        self.env.close()
        self.clean()

    def clean(self):
        output_files = [
            cwd.joinpath("testmodel_kappaexplain.lp"),
            cwd.joinpath("testmodel_kappaexplain.mps"),
        ]
        for file in output_files:
            if file.exists():
                os.remove(file)

    def test_byrows(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, expltype="ROWS")
        assert lpfile.exists()

    def test_bycols(self):
        mpsfile = cwd.joinpath("testmodel_kappaexplain.mps")
        assert not mpsfile.exists()
        kappa_explain(self.model, expltype="COLS")
        assert mpsfile.exists()

    def test_lpsubprob(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, relobjtype="LP")
        assert lpfile.exists()

    def test_qpsubprob(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, relobjtype="QP")
        assert lpfile.exists()

    def test_default(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, method="DEFAULT")
        assert lpfile.exists()

    def test_lasso(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, method="LASSO")
        assert lpfile.exists()

    def test_rls(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, method="RLS")
        assert lpfile.exists()

    def test_submatrix(self):
        lpfile = cwd.joinpath("testmodel_kappaexplain.lp")
        assert not lpfile.exists()
        kappa_explain(self.model, submatrix=True)
        assert lpfile.exists()

    #
    #   angle_explain is different, as it doesn't write a file
    #   There's a problem is any of the 3 items it returns
    #   is None
    #
    def test_angles(self):
        list1, list2, model = kappa_explain(self.model, method="ANGLES")
        assert list1 != None and list2 != None and model != None
        print("Angle test completed.")



import io
import os
import pathlib
import hashlib
import subprocess
import contextlib
from unittest import TestCase

import gurobipy as gp
from gurobi_modelanalyzer import SolCheck

here = pathlib.Path(__file__).parent
cwd = pathlib.Path(os.getcwd())


class TestSolCheck(TestCase):
    def setUp(self):
        self.env = gp.Env()
        self.afiro_model = gp.read(str(here / "dataset" / "afiro.mps"), env=self.env)
        self.misc07_model = gp.read(str(here / "dataset" / "misc07.mps"), env=self.env)

    def tearDown(self):
        self.afiro_model.close()
        self.misc07_model.close()
        self.env.close()

    def test_suboptimal_solution(self):
        m = self.afiro_model
        sol = {m.getVarByName("X01"): 78, m.getVarByName("X22"): 495}
        sc = SolCheck(m)

        captured_output = io.StringIO()
        with contextlib.redirect_stdout(captured_output):
            sc.test_sol(sol)
        self.assertIn(
            "Solution is feasible for feasibility tolerance of 1e-06",
            captured_output.getvalue(),
        )

        sc.optimize()

        for v in sol.keys():
            self.assertLess(sol[v], v.X)
        self.assertGreater(abs(sc.obj_diff), 1e-6)

    # Test an infeasible solution
    def test_infeasible_solution(self):
        m = self.misc07_model
        sol = {m.getVarByName("COL260"): 2400.5}
        sc = SolCheck(m)

        captured_output = io.StringIO()
        with contextlib.redirect_stdout(captured_output):
            sc.test_sol(sol)

        self.assertIn(
            "Model is infeasible",
            captured_output.getvalue(),
        )
        self.assertEqual(sc.Status, gp.GRB.INFEASIBLE)

        # Fix violations

        captured_output = io.StringIO()
        with contextlib.redirect_stdout(captured_output):
            sc.inf_repair()

        self.assertIn(
            "Relaxing to find smallest violation from fixed solution",
            captured_output.getvalue(),
        )
        self.assertIn(
            "Fixed values are 1.5 from a feasible solution",
            captured_output.getvalue(),
        )

        abs_sum_violations = 0

        for c in m.getConstrs():
            if abs(c._Violation) > 1e-4:
                abs_sum_violations += abs(c._Violation)

        self.assertEqual(abs_sum_violations, 1.5)


class TestSolCheckCLI(TestCase):
    def setUp(self):
        self.afirofix_md5 = "f265f0c9554f2b480092757a83e47207"

    def tearDown(self):
        output_files = [
            cwd.joinpath("afirofix.sol"),
            cwd.joinpath("misc07fix.vio"),
            cwd.joinpath("misc07fix.sol"),
            cwd.joinpath("misc07fix.ilp"),
        ]
        for file in output_files:
            if file.exists():
                os.remove(file)

    def test_suboptimal_solution(self):
        afirofix_file = cwd.joinpath("afirofix.sol")
        self.assertFalse(afirofix_file.exists())
        cmd = (
            f"gurobi_solcheck --model {str(here / 'dataset' / 'afiro.mps')} "
            + f"--sol {str(here / 'dataset' / 'afiro.sol')} --result afirofix"
        )
        result = subprocess.getoutput(cmd)
        self.assertIn(
            "Solution is feasible for feasibility tolerance of 1e-06",
            result,
        )
        self.assertIn("Difference: -5.0613", result)
        self.assertTrue(afirofix_file.exists())

        md5_ret = None
        with open(afirofix_file, "rb") as f:
            d = f.read()
            md5_ret = hashlib.md5(d).hexdigest()

        # broken on Windows, disabled for now
        # self.assertEqual(self.afirofix_md5, md5_ret)

    def test_suboptimal_json_solution(self):
        afirofix_file = cwd.joinpath("afirofix.sol")
        self.assertFalse(afirofix_file.exists())
        cmd = (
            f"gurobi_solcheck --model {str(here / 'dataset' / 'afiro.mps')} "
            + f"--sol {str(here / 'dataset' / 'afiro.json')} --result afirofix"
        )
        result = subprocess.getoutput(cmd)
        self.assertIn(
            "Solution is feasible for feasibility tolerance of 1e-06",
            result,
        )
        self.assertIn("Difference: -5.0613", result)
        self.assertTrue(afirofix_file.exists())

        md5_ret = None
        with open(afirofix_file, "rb") as f:
            d = f.read()
            md5_ret = hashlib.md5(d).hexdigest()

        # broken on Windows, disabled for now
        # self.assertEqual(self.afirofix_md5, md5_ret)

    def test_infeasible_solution(self):
        misc07fix_vio_file = cwd.joinpath("misc07fix.vio")
        self.assertFalse(misc07fix_vio_file.exists())
        cmd = (
            f"gurobi_solcheck --model {str(here / 'dataset' / 'misc07.mps')} "
            + f"--sol {str(here / 'dataset' / 'misc07.sol')} --result misc07fix"
        )
        result = subprocess.getoutput(cmd)
        self.assertTrue(misc07fix_vio_file.exists())

        # Check log for expected output
        self.assertIn(
            "Model is infeasible",
            result,
        )
        self.assertIn(
            "Solution is infeasible for feasibility tolerance of 1e-06", result
        )
        self.assertIn(
            "Relaxing to find smallest violation from fixed solution",
            result,
        )
        self.assertIn("Fixed values are 1.5 from a feasible solution", result)

    def test_infeasible_solution_infmethodV(self):
        misc07fix_sol_file = cwd.joinpath("misc07fix.sol")
        self.assertFalse(misc07fix_sol_file.exists())
        cmd = (
            f"gurobi_solcheck --model {str(here / 'dataset' / 'misc07.mps')} "
            + f"--sol {str(here / 'dataset' / 'misc07.sol')} --result misc07fix "
            + "--infmethod V "
        )
        result = subprocess.getoutput(cmd)
        self.assertTrue(misc07fix_sol_file.exists())

        # Check log for expected output
        self.assertIn("Model is infeasible", result)
        self.assertIn(
            "Relaxing to find smallest violation from fixed solution",
            result,
        )
        self.assertIn("Fixed values are 409.5 from a feasible solution", result)

    def test_infeasible_solution_infmethodI(self):
        misc07fix_ilp_file = cwd.joinpath("misc07fix.ilp")
        self.assertFalse(misc07fix_ilp_file.exists())
        cmd = (
            f"gurobi_solcheck --model {str(here / 'dataset' / 'misc07.mps')} "
            + f"--sol {str(here / 'dataset' / 'misc07.sol')} --result misc07fix "
            + "--infmethod I "
        )
        result = subprocess.getoutput(cmd)
        self.assertTrue(misc07fix_ilp_file.exists())

        # Check log for expected output
        self.assertIn("Model is infeasible", result)
        self.assertIn(
            "Solution is infeasible for feasibility tolerance of 1e-06", result
        )
        self.assertIn(
            "Computing Irreducible Inconsistent Subsystem (IIS)",
            result,
        )
        self.assertIn("IIS computed: 1 constraints, 2 bounds", result)



"""Use Gurobi to diagnose a given solution for feasibility and optimality"""

from gurobipy import GRB
import re
import gzip
import sys
import json
import itertools


def openc(fn, mode="r"):
    """Open a file with optional gzip compression"""
    return gzip.open(fn, mode + "t") if ext_check(fn, ("gz",)) else open(fn, mode)


def add_dots(fns):
    """Prepend dot to each filename in a collection (list)"""
    return map(lambda x: "." + x, fns)


def ext_check(fn, exts, compress=[]):
    """Check that a filename ends with a given list of extensions, with optional
    extensions for compression"""
    compress = [""] + list(add_dots(compress))
    return fn.lower().endswith(
        tuple(map("".join, itertools.product(add_dots(exts), compress)))
    )


class SolCheck:
    def __init__(self, model):
        self.model = model
        self.inf_method = None
        self.Status = GRB.LOADED
        self.obj_diff = None
        self.__solpat = re.compile("(\S+)\s+([+\-]?\d+\.?\d*([eE][+\-]?\d+)?)")

    def message(self, lines, pad=True):
        """Write lines using model logging"""
        if not isinstance(lines, list):
            # Embed string in a list
            lines = [lines]
        if pad:
            # Add blank lines before and after message
            lines.insert(0, "")
            lines.append("")
        for line in lines:
            self.model.message(line)

    def sol_fix(self, sol):
        """Fix variable bounds to solution values, and save original bounds"""
        self.sol = sol
        for var, val in self.sol.items():
            var._LB, var._UB = var.LB, var.UB
            var.LB, var.UB = val, val

    def sol_unfix(self):
        """Restore original variable bounds"""
        for var in self.model.getVars():
            try:
                var.LB, var.UB = var._LB, var._UB
            except AttributeError:
                pass

    def read_sol(self, fn, useZeros=False):
        """Read solution file"""
        if ext_check(fn, ("sol",), ("gz",)):
            # Handle SOL format
            sol = {}
            with openc(fn) as fp:
                for line in fp:
                    mat = self.__solpat.match(line)
                    if mat:
                        vname = mat.group(1)
                        val = float(mat.group(2))
                        sol[vname] = val
        elif ext_check(fn, ("json",), ("gz",)):
            # Handle JSON format
            with openc(fn) as fp:
                data = json.load(fp)
            sol = {v["VarName"]: v["X"] for v in data["Vars"]}
        else:
            raise RuntimeError(
                f"{fn} solution file must be either .sol(.gz) or .json(.gz)"
            )

        outsol = {}
        for vname, val in sol.items():
            if val != 0.0 or useZeros:
                var = self.model.getVarByName(vname)
                if var:
                    outsol[var] = val
                else:
                    self.message(f"Warning: {vname} not found in model", pad=False)

        return outsol

    def make_zero_sol(self):
        """Make zero solution"""
        return {v: 0.0 for v in self.model.getVars()}

    def write_vio_file(self, fn):
        """Write solution violation file"""
        with openc(fn, "w") as fp:
            fp.write(f"# Total constraint violation: {self.model.ObjVal}\n")
            for c in self.model.getConstrs():
                fp.write(f"{c.ConstrName} {c._Violation}\n")

    def write_result(self, fn):
        """Write output file"""
        if self.Status == GRB.LOADED:
            raise RuntimeError("Must test solution first")

        if self.Status == GRB.INFEASIBLE and self.inf_method == "I":
            if not ext_check(fn, ("ilp",), ("gz", "bz2", "7z", "zip")):
                fn += ".ilp"
        elif self.Status == GRB.INFEASIBLE and self.inf_method == "C":
            if not ext_check(fn, ("vio",), ("gz",)):
                fn += ".vio"
        else:
            if not ext_check(fn, ("sol",), ("gz", "bz2", "7z", "zip")):
                fn += ".sol"
        self.message([f"Writing result file {fn}"])
        if self.Status == GRB.INFEASIBLE and self.inf_method == "C":
            self.write_vio_file(fn)
        else:
            self.model.write(fn)

    def test_sol(self, sol):
        """Test a solution"""
        self.sol_fix(sol)
        self.model.optimize()
        self.Status = self.model.Status
        if self.Status == GRB.OPTIMAL:
            self.Status = GRB.SUBOPTIMAL
            feasMsg = "feasible"
        else:
            feasMsg = "infeasible"

        self.message(
            f"Solution is {feasMsg} for feasibility tolerance of {self.model.params.FeasibilityTol}"
        )

    def inf_explain(self):
        """Explain infeasible solution"""
        if self.Status == GRB.LOADED:
            raise RuntimeError("Must test solution first")

        # Force the solution into the IIS
        for var in self.sol.keys():
            var.IISLBForce = 1
            var.IISUBForce = 1

        self.model.computeIIS()
        self.inf_method = "I"

    def inf_repair(self, repairMethod="C", makeCopy=False):
        """Repair a solution or model"""
        if self.Status == GRB.LOADED:
            raise RuntimeError("Must test solution first")

        self.message("Relaxing to find smallest violation from fixed solution")
        origNumVars = self.model.NumVars

        if makeCopy:
            relax = self.model.copy()
            self.model._relax = relax
        else:
            relax = self.model

        if repairMethod == "C":  # relax constraints
            relax.feasRelax(
                0, False, None, None, None, relax.getConstrs(), relax.NumConstrs * [1]
            )
        elif repairMethod == "V":  # relax variables
            relax.feasRelax(
                0,
                False,
                list(self.sol.keys()),
                len(self.sol) * [1],
                len(self.sol) * [1],
                None,
                None,
            )
        else:
            raise RuntimeError(f"repairMethod {repairMethod} not recognized")

        relax.optimize()
        self.inf_method = repairMethod

        if repairMethod == "C":  # add attribute for constraint violations
            rval = iter(relax.getAttr("X", relax.getVars()[origNumVars:]))
            for c in self.model.getConstrs():
                c._Violation = 0.0
                if c.Sense in ("=", ">"):  # ArtP_
                    c._Violation += next(rval)
                if c.Sense in ("=", "<"):  # ArtN_
                    c._Violation -= next(rval)

        self.message(f"Fixed values are {relax.ObjVal} from a feasible solution")

    def optimize(self):
        """Optimize from test solution"""
        if self.Status == GRB.LOADED:
            raise RuntimeError("Must test solution first")

        self.message("Comparing quality with original solution")
        self.model._FixObjVal = self.model.ObjVal

        self.sol_unfix()
        self.model.optimize()

        self.obj_diff = self.model.ObjVal - self.model._FixObjVal

        # We use MIP gap even if the model is not a MIP to measure the sameness
        # of the objective
        if abs(self.obj_diff) < self.model.Params.MIPGap:
            self.Status = GRB.OPTIMAL
        else:
            self.Status = GRB.SUBOPTIMAL

        self.message(
            [
                "Objectives:",
                f"Fixed:      {self.model._FixObjVal:.4f}",
                f"Optimal:    {self.model.ObjVal:.4f}",
                f"Difference: {self.obj_diff:.4f}",
            ]
        )


# Command line code
def main_cli():
    import argparse
    import gurobipy as gp

    parser = argparse.ArgumentParser(
        prog="gurobi_solcheck", description="Check if a solution is feasible"
    )

    parser.add_argument("--model", "-m", help="Model file")
    parser.add_argument("--sol", "-s", help="Solution file to test")
    parser.add_argument(
        "--result", "-r", help="Filename where the result should be written"
    )
    parser.add_argument(
        "--testonly", action="store_true", help="Test without solution diagnosis"
    )
    parser.add_argument(
        "--infmethod",
        choices=("C", "V", "I"),
        type=str.upper,
        default="C",
        help="Method to diagnose infeasible solution (C: Repair by adjusting RHS of constraints, V: Repair by adjusting variable values, I: Explain via IIS)",
    )
    parser.add_argument(
        "--usezeros",
        "-z",
        action="store_true",
        help="Use zero values in the solution file",
    )

    args = parser.parse_args()

    if args.model:
        mfn = args.model
    else:
        import questionary as qy

        print("Entering interactive mode")
        mfn = qy.path(
            "Select a model file",
            validate=lambda fn: ext_check(
                fn, ("mps", "lp", "rew", "rlp"), ("gz", "bz2", "7z", "zip")
            ),
        ).ask()

    # Read model
    sc = SolCheck(gp.read(mfn))

    if args.model:
        sfn = args.sol
        useZeros = args.usezeros
    else:
        if qy.confirm("Use a solution file?").ask():
            sfn = qy.path(
                "Select a solution file",
                validate=lambda fn: ext_check(fn, ("sol", "json"), ("gz",)),
            ).ask()
            useZeros = not qy.confirm("Ignore zeros in solution file?").ask()
        else:
            sfn = ""

    if sfn:
        sol = sc.read_sol(sfn, useZeros)
    else:
        if args.model:
            sc.message("No solution file specified, using zero solution")
        sol = sc.make_zero_sol()

    # Check solution
    sc.test_sol(sol)

    if args.model:
        diagnose = not args.testonly
    else:
        diagnose = qy.confirm("Diagnose solution?").ask()

    if diagnose:
        if sc.Status == GRB.INFEASIBLE:
            if args.model:
                inf_method = args.infmethod
            else:
                inf_method = qy.select(
                    "Select method to diagnose infeasible solution",
                    choices=[
                        qy.Choice("Repair by adjusting RHS of constraints", "C"),
                        qy.Choice("Repair by adjusting variable values", "V"),
                        qy.Choice("Explain via IIS", "I"),
                    ],
                ).ask()

            if inf_method == "I":
                sc.inf_explain()
            else:
                sc.inf_repair(inf_method)
        else:
            sc.optimize()

        if args.model:
            rfn = args.result
        else:
            if qy.confirm("Write result file?").ask():
                rfn = qy.path("Select a result file").ask()
            else:
                rfn = ""

        if rfn:
            sc.write_result(rfn)

    sys.exit(sc.Status)



import contextlib
import datetime
import gurobipy as gp
from io import StringIO
import os
import sys
from timeit import default_timer

import gurobi_modelanalyzer.common as common
import gurobi_modelanalyzer.constraint_analyzer as constraint_analyzer
import gurobi_modelanalyzer.file_analyzer as file_analyzer
import gurobi_modelanalyzer.objective_function_analyzer as objective_function_analyzer
import gurobi_modelanalyzer.variable_analyzer as variable_analyzer


def run(model_file: str, analyzers=[]):
    # Uncompress model file if compressed
    fileType, compression = file_analyzer.get_file_type(model_file)

    if fileType is None:
        print(f"Error: {model_file} has unsupported file type")
        return 1

    # Initialize data dictionary
    data = {
        "fileName": os.path.basename(model_file),
        "fileType": fileType.upper(),
        "reportTimestamp": common.utc_date_formatted(datetime.datetime.utcnow()),
        "gurobiVersion": "%d.%d.%d" % gp.gurobi.version(),
        "fileSize": os.stat(model_file).st_size,
        "fileSizeUncompressed": 0,
        "readerSuccess": False,
        "readerWarnings": False,
        "readerError": False,
    }

    # Read model into Gurobi
    try:
        print(f"Reading model file '{model_file}' ...")
        with elapsed_timer() as elapsed:
            with capture_output() as out:
                m = gp.read(model_file)
    except:
        data["readerError"] = True

    finally:
        data["readerOutput"] = [line.rstrip("\n") for line in out if len(line) > 0]
        data["readerTime"] = elapsed()

    if data["readerError"]:
        print("ERROR reading model file")
        return 1

    if compression is not None:
        data["isCompressed"] = True

        try:
            file_analyzer.process_compressed_model(compression, model_file, data)

        except Exception as e:
            print("ERROR while trying do decompress model file")
            print(e)
            return 1
    else:
        data["isCompressed"] = False
        file_analyzer.process_uncompressed_model(model_file, data)

    # Check for warnings
    if any("Warning" in line for line in data["readerOutput"]):
        data["readerWarnings"] = True
    else:
        data["readerSuccess"] = True

    process_model_type(m, data)
    if "objective" in analyzers:
        objective_variables = objective_function_analyzer.process_objective(m, data)

    if "constraint" in analyzers:
        constraint_variables = constraint_analyzer.process_constraints(m, data)

    if "variable" in analyzers:
        variable_analyzer.process_variable_types(
            m, data, objective_variables, constraint_variables
        )

    if "bounds" in analyzers:
        constraint_analyzer.process_bounds(m, data)

    if "coefficients" in analyzers:
        constraint_analyzer.process_coefficients(m, data)

    if "rhs" in analyzers:
        constraint_analyzer.process_rhs(m, data)

    # Write matrix plot image and put final dimensions into data
    # write_matrix_plot(m, 1920, 1080, args.outputprefix + ".png", data)

    return data


def process_model_type(m, data):
    data["isMIP"] = bool(m.IsMIP)
    data["isQP"] = bool(m.IsQP)
    data["isQCP"] = bool(m.IsQCP)

    part1_short = "MI" if data["isMIP"] == 1 else ""
    part1_long = "Mixed-Integer " if data["isMIP"] == 1 else ""

    part2_short = "Q" if data["isQP"] == 1 or data["isQCP"] == 1 else ""
    part2_long = "Quadratic " if data["isQP"] and not data["isQCP"] else ""

    part3_short = "C" if data["isQCP"] else ""
    part3_long = "Quadratically Constrained " if data["isQCP"] else ""

    part4_short = (
        "LP" if part1_short == "" and part2_short == "" and part3_short == "" else "P"
    )
    part4_long = (
        "Linear Program"
        if part1_long == "" and part2_long == "" and part3_long == ""
        else "Program"
    )

    data["modelTypeShort"] = part1_short + part2_short + part3_short + part4_short
    data["modelTypeLong"] = part1_long + part2_long + part3_long + part4_long
    print("Model Type: %s" % data["modelTypeShort"])


@contextlib.contextmanager
def capture_output():
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [StringIO(), StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


@contextlib.contextmanager
def elapsed_timer():
    start = default_timer()
    elapser = lambda: default_timer() - start
    yield lambda: elapser()
    end = default_timer()
    elapser = lambda: end - start



from unittest import TestCase
import pathlib

import gurobi_modelanalyzer.analyzer as analyzer


class TestAnalyzer(TestCase):
    def setUp(self) -> None:
        here = pathlib.Path(__file__).parent
        self.file_path = str(here / "dataset" / "glass4.mps")

    def test_run(self):
        data = analyzer.run(self.file_path)
        print("Done")



from collections import defaultdict
import gurobipy as gp
import math

from gurobi_modelanalyzer import common


def get_rhs_frequencies(model, base=10):
    rhs_count = defaultdict(int)
    rhs_vals = []
    result = []

    for constr in model.getConstrs():
        if constr.RHS != 0:
            rhs_vals.append(constr.RHS)

    result = common.get_vector_frequencies(rhs_vals, base)

    return result


def get_a_frequencies(model, base, topx):
    int_coeff_count = defaultdict(int)
    cont_coeff_count = defaultdict(int)
    range_count = defaultdict(int)
    rows = []

    for constr in model.getConstrs():
        row = model.getRow(constr)

        min_exponent = float("inf")
        min_abs_coeff = float("inf")
        max_exponent = -float("inf")
        max_abs_coeff = -float("inf")

        # Loop through all terms, count single coefficient exponents and total range row
        for i in range(row.size()):
            coeff = row.getCoeff(i)
            if coeff != 0:
                exponent = int(math.floor(math.log(abs(coeff), base)))

                if row.getVar(i).VType == "C":
                    cont_coeff_count[exponent] += 1
                else:
                    int_coeff_count[exponent] += 1

                if exponent < min_exponent:
                    min_exponent = exponent
                    min_var = row.getVar(i)
                    min_abs_coeff = abs(coeff)
                if exponent > max_exponent:
                    max_exponent = exponent
                    max_var = row.getVar(i)
                    max_abs_coeff = abs(coeff)

        # Record range for non-empty constraints
        if min_exponent < float("inf"):
            range_count[max_exponent - min_exponent] += 1

            if topx > 0:
                rows.append(
                    {
                        "constr_name": constr.ConstrName,
                        "range": max_exponent - min_exponent,
                        "min_exponent": min_exponent,
                        "min_abs_coeff": min_abs_coeff,
                        "min_var": min_var.VarName,
                        "min_var_type": min_var.VType,
                        "max_exponent": max_exponent,
                        "max_abs_coeff": max_abs_coeff,
                        "max_var": max_var.VarName,
                        "max_var_type": max_var.VType,
                    }
                )

    cont_coeff_result = [
        [exponent, cont_coeff_count[exponent]]
        for exponent in range(
            min(cont_coeff_count.keys()), max(cont_coeff_count.keys()) + 1
        )
    ]
    int_coeff_result = [
        [exponent, int_coeff_count[exponent]]
        for exponent in range(
            min(int_coeff_count.keys()), max(int_coeff_count.keys()) + 1
        )
    ]
    range_result = [
        [exponent, range_count[exponent]]
        for exponent in range(min(range_count.keys()), max(range_count.keys()) + 1)
    ]

    # Get the TOP x results
    if topx > 0:
        topx_results = sorted(rows, key=lambda k: k["range"], reverse=True)[:topx]
    else:
        topx_results = []

    return cont_coeff_result, int_coeff_result, range_result, topx_results


def process_bounds(model, data):
    print("Processing bounds...")

    for i in [2, 10]:
        lb, ub = common.get_variable_bounds(model, data, i)
        data["varLowerBoundsLog" + str(i)] = lb
        data["varUpperBoundsLog" + str(i)] = ub


def process_coefficients(model, data):
    print("Processing coefficients...")

    top_ranges = []
    top_x = 10

    for i in [2, 10]:
        (
            data["objContinuousExponentsLog" + str(i)],
            data["objIntegerExponentsLog" + str(i)],
        ) = common.get_obj_c_frequencies(model, i)

        (
            a_cont_coeff_count,
            a_int_coeff_count,
            a_range_count,
            top_ranges,
        ) = get_a_frequencies(model, i, top_x)

        data["linearConstraintsContinuousExponentsLog" + str(i)] = a_cont_coeff_count
        data["linearConstraintsIntegerExponentsLog" + str(i)] = a_int_coeff_count
        data["linearConstraintsRowRangesLog" + str(i)] = a_range_count
        data["topSingleRowRangesLog" + str(i)] = top_ranges


def process_rhs(model, data):
    print("Processing RHS values...")

    for i in [2, 10]:
        data["rhsExponentsLog" + str(i)] = get_rhs_frequencies(model, i)


def process_linear_constraints(model, data):
    print("Processing model linear constraints...")

    constraint_variables = set()
    redundant_constraints = []
    infeasible_constraints = []

    # Equality constraint
    counter_equality = 0

    # Inequality constraint
    counter_inequality = 0

    # Empty constraint
    counter_empty = 0

    # Bound constraint
    counter_bound = 0

    # Integer and continuous variables
    counter_mixed = 0

    # Only (>= 2) continuous variables
    counter_con_only = 0

    # Only (>= 2) binary variables
    counter_bin_only = 0

    # Only (>= 2) integer variables (including binaries)
    counter_int_only = 0

    # sum xi >= 1 (xi bin)
    counter_set_covering = 0

    # sum xi = 1 (xi bin)
    counter_set_partitioning = 0

    # sum xi <= 1 (xi bin)
    counter_set_packing = 0

    # sum xi <= c (xi bin, c integer != 1)
    counter_max_binsum = 0

    # sum xi = c (xi bin, c integer != 1)
    counter_binsum = 0

    # sum xi >= c (xi bin, c integer != 1)
    counter_min_binsum = 0

    # sum c_i x_i <= d (xi bin, ci, d positive integer)
    counter_bin_knapsack = 0

    # sum c_i x_i <= d (xi integer, ci, d positive integer)
    counter_int_knapsack = 0

    # Redundant constraints (i.e. constraint is always satisfied)
    counter_redundant = 0

    # Infeasible constraints (i.e. constraint can never be satisfied)
    counter_infeasible = 0

    # Loop through all linear constraints
    for constraint in model.getConstrs():
        if constraint.Sense == gp.GRB.EQUAL:
            counter_equality += 1
        else:
            counter_inequality += 1

        row = model.getRow(constraint)

        rhs = constraint.RHS - row.getConstant()
        is_rhs_integer = abs(rhs - round(rhs)) < 1e-10

        # Empty constraints
        if row.size() == 0:
            counter_empty += 1

        # Empty or bound constraint
        elif row.size() == 1:
            counter_bound += 1

        min_coefficient = gp.GRB.INFINITY
        max_coefficient = -gp.GRB.INFINITY

        num_binvars = 0
        num_intvars = 0

        is_row_coeff_integer = True
        is_row_all_positive = True
        is_row_all_negative = True

        lhs_lb = row.getConstant()
        lhs_ub = row.getConstant()

        for i in range(row.size()):
            var = row.getVar(i)
            coefficient = row.getCoeff(i)

            if lhs_lb > -gp.GRB.INFINITY:
                lhs_lb = max(
                    -gp.GRB.INFINITY, lhs_lb + common.min_value(var, coefficient)
                )

            if lhs_ub < gp.GRB.INFINITY:
                lhs_ub = min(
                    gp.GRB.INFINITY, lhs_ub + common.max_value(var, coefficient)
                )

            if coefficient != 0:
                constraint_variables.add(var)

            min_coefficient = min(min_coefficient, coefficient)
            max_coefficient = max(max_coefficient, coefficient)

            if (coefficient - round(coefficient)) > 1e-10:
                is_row_coeff_integer = False

            if coefficient < 0:
                is_row_all_positive = False

            if coefficient > 0:
                is_row_all_negative = False

            if common.get_vtype(var) == gp.GRB.BINARY:
                num_binvars += 1
            elif common.get_vtype(var) == gp.GRB.INTEGER:
                num_intvars += 1

        # Check if constraint is redundant
        if (
            constraint.Sense == gp.GRB.LESS_EQUAL
            and lhs_ub <= constraint.RHS
            or constraint.Sense == gp.GRB.GREATER_EQUAL
            and lhs_lb >= constraint.RHS
            or constraint.Sense == gp.GRB.EQUAL
            and lhs_lb == constraint.RHS
            and lhs_ub == constraint.RHS
        ):
            counter_redundant += 1
            redundant_constraints.append(constraint)

        # Check if constraint is infeasible
        if (
            constraint.Sense == gp.GRB.LESS_EQUAL
            and lhs_lb > constraint.RHS
            or constraint.Sense == gp.GRB.GREATER_EQUAL
            and lhs_ub < constraint.RHS
            or constraint.Sense == gp.GRB.EQUAL
            and (lhs_ub < constraint.RHS or lhs_lb > constraint.RHS)
        ):
            counter_infeasible += 1
            infeasible_constraints.append(constraint)

        # Continue only for constraints that contain at least two variables
        if row.size() < 2:
            continue

        # Int-Only / Cont-Only / Mixed
        if num_binvars + num_intvars == 0:
            counter_con_only += 1
        elif num_binvars + num_intvars == row.size():
            counter_int_only += 1
        else:
            counter_mixed += 1

        # All variables binary?
        if num_binvars == row.size():
            counter_bin_only += 1

            sense = constraint.Sense

            if rhs < 0:
                factor = -1
                if sense == gp.GRB.LESS_EQUAL:
                    sense = gp.GRB.GREATER_EQUAL
                if sense == gp.GRB.GREATER_EQUAL:
                    sense = gp.GRB.LESS_EQUAL
            else:
                factor = 1

            # All coefficients 1?
            if (
                abs(min_coefficient - max_coefficient) < 1e-10
                and abs(min_coefficient - factor) < 1e-10
            ):
                # RHS = 1?
                if abs(rhs - factor) < 1e-10:
                    if constraint.Sense == gp.GRB.EQUAL:
                        counter_set_partitioning += 1
                    elif constraint.Sense == gp.GRB.LESS_EQUAL:
                        counter_set_packing += 1
                    else:
                        counter_set_covering += 1
                    continue

                # RHS integer?
                elif is_rhs_integer:
                    if constraint.Sense == gp.GRB.EQUAL:
                        counter_binsum += 1
                    elif constraint.Sense == gp.GRB.LESS_EQUAL:
                        counter_max_binsum += 1
                    else:
                        counter_min_binsum += 1
                    continue

            # Binary knapsack constraints (all variables binary, all coefficients + RHS integer)
            if (
                is_row_coeff_integer
                and (
                    (is_row_all_positive and rhs > 0)
                    or (is_row_all_negative and rhs < 0)
                )
                and is_rhs_integer
            ):
                counter_bin_knapsack += 1

        # All variables integer?
        if num_intvars == row.size():
            # Knapsack constraints (all variables binary, all coefficients + RHS integer)
            if (
                is_row_coeff_integer
                and (
                    (is_row_all_positive and rhs > 0)
                    or (is_row_all_negative and rhs < 0)
                )
                and is_rhs_integer
            ):
                counter_int_knapsack += 1

    # TOP 5 redundant constraints
    redundant_constraints = (
        sorted(
            redundant_constraints,
            key=lambda constraint: model.getRow(constraint).size(),
        )
    )[:5]
    data["redundantConstraints"] = [
        constraint.ConstrName for constraint in redundant_constraints
    ]

    # TOP 5 infeasible constraints
    infeasible_constraints = (
        sorted(
            infeasible_constraints,
            key=lambda constraint: model.getRow(constraint).size(),
        )
    )[:5]
    data["infeasibleConstraints"] = [
        constraint.ConstrName for constraint in infeasible_constraints
    ]

    data["numEqConstrs"] = counter_equality
    data["numIneqConstrs"] = counter_inequality

    data["numEmptyConstrs"] = counter_empty
    data["numBndConstrs"] = counter_bound
    data["numRedundantConstrs"] = counter_redundant
    data["numInfeasibleConstrs"] = counter_infeasible

    data["numContConstrs"] = counter_con_only
    data["numIntConstrs"] = counter_int_only
    data["numBinConstrs"] = counter_bin_only
    data["numMixedConstrs"] = counter_mixed

    data["numSetCovConstrs"] = counter_set_covering
    data["numSetPartConstrs"] = counter_set_partitioning
    data["numSetPackConstrs"] = counter_set_packing

    data["numMaxBinSumConstrs"] = counter_max_binsum
    data["numMinBinSumConstrs"] = counter_min_binsum
    data["numBinSumConstrs"] = counter_binsum

    data["numBinKnapsackConstrs"] = counter_bin_knapsack
    data["numIntKnapsackConstrs"] = counter_int_knapsack

    return constraint_variables


def process_quadratic_constraints(model, data):
    #
    #   TODO: identify infeasible and redundant QCs.
    #   Not sure worth doing for QCs, but what about semi cont. and
    #   semi int variables?
    #

    print("Processing model quadratic constraints...")

    quadpart_variables = set()
    linpart_variables = set()

    redundant_constraints = []
    infeasible_constraints = []

    # Equality constraint
    counter_equality = 0

    # Inequality constraint
    counter_inequality = 0

    # Empty constraint
    counter_empty = 0

    # Bound constraint
    counter_bound = 0

    # Integer and continuous variables
    if model.NumIntVars + model.NumBinVars >= 1:
        counter_mixed = 0
    else:
        counter_mixed = -1

    # Only (>= 2) continuous variables

    if model.NumVars - model.NumIntVars - model.NumBinVars >= 2:
        counter_cont_only = 0
    else:
        counter_cont_only = -1  # Can't have bilinear terms with 2 cont. vars

    # Only (>= 2) binary variables
    if model.NumBinVars >= 2:
        counter_bin_only = 0
    else:
        counter_bin_only = -1  # Can't have any binary bilinear terms

    # Only (>= 2) integer variables (including binaries)
    if model.NumIntVars >= 1:
        counter_int_only = 0
    else:
        counter_int_only = -1  # Can't have any integer bilinear terms

    # No linear terms in the QC (bot not totally empty)

    counter_quad_only = 0

    # Empty quadratic part; should have made it a linear constraint

    counter_lin_only = 0

    # At least one binary in every quadratic term of the QC

    if model.NumBinVars >= 1:
        counter_common_bin = 0
    else:
        counter_common_bin = -1

    # Anything else regarding QC term composition

    counter_other = 0

    # Loop through all quadratic constraints

    BIN_ONLY = 1
    INT_ONLY = 2
    CONT_ONLY = 4
    MIXED = 8
    for qconstraint in model.getQConstrs():
        qcrow = model.getQCRow(qconstraint)
        row = qcrow.getLinExpr()
        qlen = qcrow.size()
        linlen = row.size()
        bittype = 0
        qclhs_lb = row.getConstant()
        qclhs_ub = row.getConstant()

        # Loop quadratic terms

        if qlen == 1:  #  Check for QC singleton that is a simple bound
            if qcrow.getCoeff(0) != 0:
                v1 = qcrow.getVar1(0)
                v2 = qcrow.getVar2(0)
                if v1.VarName == v2.VarName and linlen == 0:
                    counter_bound += 1
                if v1.vType == gp.GRB.BINARY:
                    counter_common_bin += 1
                    if v2.vType == gp.GRB.BINARY:
                        bittype != BIN_ONLY
                    else:
                        bittype != MIXED
                elif v2.vType == gp.GRB.BINARY:
                    counter_common_bin += 1
                    bittype != MIXED
                else:  # QC singleton of two continuous variables
                    bittype != CONT_ONLY
            elif linlen == 0:
                counter_empty += 1
            else:  #  Empty quad part with nonempty linear part
                counter_lin_only += 1
        else:  # Multiple QC terms; count all pairwise combos
            binvardict = defaultdict(int)  # for common binary factors
            for j in range(qlen):
                if qcrow.getCoeff(j) != 0:
                    v1 = qcrow.getVar1(j)
                    v2 = qcrow.getVar2(j)
                    quadpart_variables.add(v1)
                    quadpart_variables.add(v2)
                    if counter_cont_only >= 0:
                        if v1.Vtype == gp.GRB.CONTINUOUS:
                            if v2.VType == gp.GRB.CONTINUOUS:
                                bittype |= CONT_ONLY
                            else:
                                bittype |= MIXED
                                if v2.Vtype == gp.GRB.BINARY:
                                    binvardict[v2.VarName] += 1
                            continue
                        elif v2.Vtype == gp.GRB.CONTINUOUS:
                            bittype |= MIXED
                            if v2.Vtype == gp.GRB.BINARY:
                                binvardict[v2.VarName] += 1
                            continue
                    if counter_bin_only >= 0:
                        if v1.Vtype == gp.GRB.BINARY and v2.VType == gp.GRB.BINARY:
                            bittype |= BIN_ONLY
                            binvardict[v1.VarName] += 1
                            if v1.VarName != v2.VarName:
                                binvardict[v1.VarName] += 1
                            continue
                    if counter_int_only >= 0:
                        if v1.Vtype == gp.GRB.INTEGER and v2.VType == gp.GRB.INTEGER:
                            bittype |= INT_ONLY
                            continue
            if len(binvardict) == qlen:
                counter_common_bin += 1

        # Finished processing this QC; update the counts of the different
        # types of bilinear terms

        if bittype == BIN_ONLY:
            counter_bin_only += 1
        elif bittype == INT_ONLY:
            counter_int_only += 1
        elif bittype == CONT_ONLY:
            counter_cont_only += 1
        elif bittype == MIXED:
            counter_mixed += 1
        else:  # Anything that doesn't fit one of the above QC types
            counter_other += 1

        # Loop linear terms.   Record contribution of linear part of QC
        # to the QC inf and sup (any constant term already recorded at
        # initialization of inf and sup).

        for i in range(linlen):
            coeff = row.getCoeff(i)
            if coeff != 0:
                var = row.getVar(i)
                linpart_variables.add(var)
                if qclhs_lb > -gp.GRB.INFINITY:
                    t = common.min_value(var, coeff)
                    if t != -gp.GRB.INFINITY:
                        qclhs_lb += t
                    else:
                        qclhs_lb = -gp.GRB.INFINITY

                if qclhs_ub < gp.GRB.INFINITY:
                    t = common.max_value(var, coeff)
                    if t != gp.GRB.INFINITY:
                        qclhs_ub += t
                    else:
                        qclhs_ub = gp.GRB.INFINITY

        if qconstraint.QCSense == gp.GRB.EQUAL:
            counter_equality += 1
        else:
            counter_inequality += 1

        if qlen == 0:
            if linlen == 0:
                counter_empty += 1
            else:
                counter_lin_only += 1
        else:
            if linlen == 0:
                counter_quad_only += 1
    #
    #   Final results.  Counters that are at -1 were flagged as
    #   not needing counting due to the numbers of different types of
    #   variables in the model (e.g. you can't have constraints involving
    #   bilinear terms of continuous variables when the model is all binary,
    #   so flag that and don't bother counting in the inner loop across the QC
    #
    data["numQCEqConstraints"] = counter_equality
    data["numQCInEqConstraints"] = counter_inequality
    data["numQCEmptyConstrs"] = counter_empty
    data["numQCBndConstrs"] = counter_bound
    if counter_cont_only == -1:
        counter_cont_only = 0
    data["numQCContConstrs"] = counter_cont_only
    data["numQCIntConstrs"] = counter_int_only
    if counter_bin_only == -1:
        counter_bin_only = 0
    data["numQCBinConstrs"] = counter_bin_only
    if counter_mixed == -1:
        counter_mixed = 0
    data["numQCMixedConstrs"] = counter_mixed
    data["numQCOtherConstrs"] = counter_other
    data["numQCQuadOnlyConstrs"] = counter_quad_only
    data["numQCLinOnlyConstrs"] = counter_lin_only
    data["numQCEmptyConstrs"] = counter_empty
    if counter_common_bin == -1:
        counter_common_bin = 0
    data["numQCBinFactorConstrs"] = counter_common_bin

    return quadpart_variables, linpart_variables



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


project = "Gurobi Model Analyzer"
copyright = "2024, Gurobi Optimization"
author = "Gurobi Optimization"
html_title = "Gurobi Model Analyzer"

html_theme = "gurobi_sphinxtheme"
html_theme_options = {
    "version_warning": False,
    "feedback_banner": False,
    "construction_warning": False,
}

html_favicon = "https://www.gurobi.com/favicon.ico"

# -- Warning banner while in beta -------------------------------------------
# rst_prolog = """.. warning::
#    This code is in a pre-release state. It may not be fully functional and breaking changes
#    can occur without notice.
# """

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.extlinks",
    "sphinxcontrib.bibtex",
]

bibtex_bibfiles = ["bib_illcond.bib"]

extlinks_detect_hardcoded_links = True
extlinks = {
    "issue": ("https://github.com/Gurobi/gurobi-machinelearning/issues/%s", "issue %s"),
    "gurobipy": (
        "https://www.gurobi.com/documentation/current/refman/py_%s.html",
        "gurobipy %s",
    ),
    "pypi": ("https://pypi.org/project/%s/", "%s"),
}


templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]



from collections import defaultdict
import gurobipy as gp

import gurobi_modelanalyzer.common as common


def process_objective(m, data):
    print("Processing objective function...")

    obj = m.getObjective()
    obj_variables = set()
    count_obj_coeffs_decimal_digits = defaultdict(int)
    count_positive_coeffs = 0
    count_linear_vars = 0
    count_pwl_vars = 0
    count_pwl_convex_vars = 0
    count_pwl_nonconvex_vars = 0
    count_quad_terms = 0

    # Quadratic terms
    if type(obj) is gp.QuadExpr:
        count_quad_terms = obj.size()
        for i in range(obj.size()):
            coeff = obj.getCoeff(i)
            if coeff > 0:
                count_positive_coeffs += 1
            count_obj_coeffs_decimal_digits[common.count_decimal_digits(coeff)] += 1
            obj_variables.add(obj.getVar1(i))
            obj_variables.add(obj.getVar2(i))
            # TODO: Include bounds for quadratic terms in obj_lb and obj_ub

        obj = obj.getLinExpr()

    count_quad_vars = len(obj_variables)
    obj_lb = obj.getConstant()
    obj_ub = obj.getConstant()

    # Linear / PWL terms
    for var in m.getVars():
        if var.Obj == 0:
            pwl = m.getPWLObj(var)
            if len(pwl) > 0:
                count_pwl_vars += 1
                obj_variables.add(var)
                if var.PWLObjCvx == 0:
                    count_pwl_nonconvex_vars += 1
                else:
                    count_pwl_convex_vars += 1
        else:
            count_linear_vars += 1
            obj_variables.add(var)
            if var.Obj > 0:
                count_positive_coeffs += 1

            if obj_lb > -gp.GRB.INFINITY:
                obj_lb = max(-gp.GRB.INFINITY, obj_lb + common.min_value(var, var.Obj))

            if obj_ub < gp.GRB.INFINITY:
                obj_ub = min(gp.GRB.INFINITY, obj_ub + common.max_value(var, var.Obj))

            count_obj_coeffs_decimal_digits[common.count_decimal_digits(var.Obj)] += 1

    obj_coeffs_decimal_digits_result = [
        [length, count_obj_coeffs_decimal_digits[length]]
        for length in range(
            min(count_obj_coeffs_decimal_digits.keys()),
            max(count_obj_coeffs_decimal_digits.keys()) + 1,
        )
    ]

    data["objLB"] = obj_lb
    data["objUB"] = obj_ub

    data["objDecimalDigits"] = obj_coeffs_decimal_digits_result
    data["numObjTotalVars"] = len(obj_variables)
    data["numObjVars"] = count_linear_vars
    data["numPWLObj"] = count_pwl_vars
    data["numPWLObjCVars"] = count_pwl_convex_vars
    data["numPWLObjNCVars"] = count_pwl_nonconvex_vars
    data["numQObjTerms"] = count_quad_terms
    data["numQObjVars"] = count_quad_vars

    data["numObjIntVars"] = len(
        [var for var in obj_variables if common.get_vtype(var) != gp.GRB.CONTINUOUS]
    )
    data["numObjCoeffs"] = sum(count_obj_coeffs_decimal_digits.values())
    data["numObjIntCoeffs"] = count_obj_coeffs_decimal_digits[0]
    data["numObjPosCoeffs"] = count_positive_coeffs

    if m.ModelSense == 1:
        data["isMin"] = True
    else:
        data["isMin"] = False

    assert count_pwl_vars == count_pwl_convex_vars + count_pwl_nonconvex_vars
    return obj_variables



import gurobipy as gp
from collections import defaultdict
import math


def get_vtype(var):
    if var.VType == gp.GRB.INTEGER and var.LB >= 0 and var.UB <= 1:
        return gp.GRB.BINARY
    else:
        return var.VType


def max_value(variable, coefficient):
    """
    Calculates the maximum attainable value for a variable. Either the upper bound or infinity.
    :param variable: The variable
    :param coefficient: The coefficient with which the upper bound should multiplied
    :return: The maximum value
    """
    if coefficient == 0:
        return 0
    elif coefficient > 0:
        return min(gp.GRB.INFINITY, variable.UB * coefficient)
    else:
        return min(gp.GRB.INFINITY, variable.LB * coefficient)


def min_value(variable, coefficient):
    """
    Calculates the minimum attainable value for a variable. Either the upper bound or infinity.
    :param variable: The variable
    :param coefficient: The coefficient with which the lower bound should multiplied
    :return: The minimum value
    """
    if coefficient == 0:
        return 0
    elif coefficient > 0:
        return max(-gp.GRB.INFINITY, variable.LB * coefficient)
    else:
        return max(-gp.GRB.INFINITY, variable.UB * coefficient)


def numfinite(vals):
    """
    Counts the number of elements in an array take on finite values
    :param vals:  The array of values to check
    :return: True or False
    """
    count = 0
    for v in vals:
        if v > -gp.GRB.INFINITY and v < gp.GRB.INFINITY:
            count += 1
    return count


def isfinite(vals):
    """
    Checks if all elements in an array take on finite values
    by calling the numfinite routine to count the number of finite
    values.
    :param vals:  The array of values to check
    :return: True or False
    """

    return numfinite(vals) == len(vals)


def qterm_min_value(var1, var2, qcoeff):
    """
    Calculates the minimum attainable value for a single quadratic term
    More complicated than the corresponding min_value routine for linear
    terms, as we need to consider 4 lower/upper bound combinations for the
    two variables in the quadratic term.
    :param var1, var2: The two variables comprising the quadratic term
    :param qcoeff: The coefficient of the quadratic term
    :return: The minimum value of the quadratic term
    """
    if qcoeff == 0:
        return 0
    #   TODO: Handle the case of infeasible bounds where at least one
    #   lower bound exceeds the corresponding upper bound, regardless of
    #   violation amount.
    #    import pdb; pdb.set_trace()

    l1 = var1.LB
    l2 = var2.LB
    u1 = var1.UB
    u2 = var2.UB
    vallist = []

    """
    Start with the easy cases that are similar to linear terms, i.e.
    none of the 4 bounds have opposite signs.
    """

    if l1 >= 0 and l2 >= 0:
        if u1 >= gp.GRB.INFINITY or u2 >= gp.GRB.INFINITY:
            vallist.extend([l1 * l2, gp.GRB.INFINITY])
        else:  # all 4 bounds are finite
            vallist.extend([l1 * u2, l2 * u1, l1 * l2, u1 * u2])
    elif u1 <= 0 and u2 <= 0:
        if l1 <= -gp.GRB.INFINITY or l2 <= -gp.GRB.INFINITY:
            vallist.extend([u1 * u2, gp.GRB.INFINITY])
        else:  # all 4 bounds are finite
            vallist.extend([l1 * u2, l2 * u1, l1 * l2, u1 * u2])

    else:
        """
        Now consider the harder cases that are probably rare in practice,
        where not all 4 variable bounds are nonnegative or nonpositive.
        Need to consider all 4 possible variable bound products, and handle
        cases of one or more infinite bounds separately.
        """
        bndlist = [l1, l2, u1, u2]
        lblist = [l1, l2]
        ublist = [u1, u2]
        minval = gp.GRB.INFINITY
        maxval = -gp.GRB.INFINITY
        if isfinite(bndlist):  # Easy; all bilinear products finite
            vallist = [l1 * u2, l2 * u1, l1 * l2, u1 * u2]
        else:
            """
            At least one bound at +-infinity.  Rather than dealing with
            all the possible different cases, there really are only a handful
            of cases to consider that we can categorize based mostly on the
            number of infinite bounds among the list of 4 bounds for var1
            and var2
            """
            ninfbds = len(bndlist) - numfinite(bndlist)
            if ninfbds >= 3:
                """
                At least one xi*xj term at +inf and at least one at -inf,
                so we know the max and min bound products without much work
                """
                vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
            elif ninfbds == 2:
                """
                We have one special case here.   If var1 or var2 are free,
                then at least one xi*xj term is +inf and at least one is
                -inf regardless of the signs of the lower and upper bound
                of the other variable.
                """
                if numfinite([l1, u1]) == 0 or numfinite([l2, u2]) == 0:
                    vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                elif numfinite([l1, l2]) == 0:
                    """
                    Both lower bounds negative infinite, so both upper bounds
                    finite.   Min and Max xi*xj terms depend on signs of
                    the finite bounds.
                    """
                    if u1 >= 0 or u2 >= 0:
                        vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                    else:  # u1, u2 nonpositive; their product is the min xi*xj
                        vallist.extend([u1 * u2, gp.GRB.INFINITY])
                elif numfinite([u1, u2]) == 0:
                    """
                    Both upper bounds infinite, so both lower bounds
                    finite.   Min and Max xi*xj terms depend on signs of
                    the finite bounds.
                    """
                    if l1 <= 0 or l2 <= 0:
                        vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                    else:  # l1, l2 positive; their product is the min xi*xj
                        vallist.extend([l1 * l2, gp.GRB.INFINITY])
                else:
                    """
                    One lower bound is -infinite and one upper bound is
                    infinite for the other variable (we handled the
                    free variable case previously).  If finite bounds are
                    both nonnegative or nonpositive, then min and max
                    xi*xj values are -inf and +inf.   Otherwise, need to
                    consider different locations of finite bounds of
                    opposite sign.
                    """
                    if numfinite([l1, u2]) == 2:
                        if l1 <= 0 and u2 <= 0 or l1 >= 0 and u2 >= 0:
                            vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                        else:
                            if u1 >= 0 and u2 >= 0 and l1 <= 0 and l2 <= 0:
                                vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                            else:
                                vallist.extend([-gp.GRB.INFINITY, l1 * u2])
                    else:  # l2 and u1 finite
                        if l2 <= 0 and u1 <= 0 or l2 >= 0 and u1 >= 0:
                            vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                        else:
                            if u1 >= 0 and u2 >= 0 and l1 <= 0 and l2 <= 0:
                                vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                            else:  # l2 >= 0 and u1 <= 0
                                vallist.extend([-gp.GRB.INFINITY, l2 * u1])
            else:
                """
                All that remains is the case when one bound is +-infinity. We've already
                handled the two cases of l1 and l2 nonnegative and u1 and u2 nonpositive.
                Find the +- infinite bound, then look at the signs of the two bounds that
                have products with the infinite bound to determine the possible products
                of pairs of variable bounds
                """
                if numfinite([l1, u1]) == 2:  # x2 has the +- infinite bound
                    if l1 <= 0 and u1 >= 0:
                        vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                    else:  # l1 and u1 have the same sign
                        infval = math.copysign(1, u1) * gp.GRB.INFINITY
                        if isfinite([l2]):  # u2 infinite
                            vallist.extend([l1 * l2, l2 * u1, infval])
                        else:  # l2 -infinite
                            vallist.extend([l1 * u2, u1 * u2, -infval])

                else:  # x1 has the +- infinite bound
                    if l2 <= 0 and u2 >= 0:
                        vallist.extend([-gp.GRB.INFINITY, gp.GRB.INFINITY])
                    else:  # l2 and u2 have the same sign
                        infval = math.copysign(1, u2) * gp.GRB.INFINITY
                        if isfinite([l1]):  # u1 infinite
                            vallist.extend([l1 * l2, l1 * u2, infval])
                        else:  # l1 -infinite
                            vallist.extend([u1 * u2, l2 * u1, -infval])

    """
    The heavy lifting is done.   Just need to look at min and max
    values of the at most 4 possible different products of the two variable bounds
    to determine min quadratic term value
    """
    minval = min(vallist)
    maxval = max(vallist)
    if qcoeff > 0:
        if isfinite([minval]):
            return qcoeff * minval
        else:
            return minval
    else:
        if isfinite([maxval]):
            return qcoeff * maxval
        else:
            return -maxval


# TODO: Need a qterm_max_value.


def utc_date_formatted(dt):
    """
    Format the date into a string
    :param dt: The date
    :return: The string
    """
    weekday = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"][dt.weekday()]
    month = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ][dt.month - 1]
    return f"{weekday}, {dt.day:02d} {month} {dt.year:04d} {dt.hour:02d}:{dt.minute:02d}:{dt.second:02d} GMT"


def count_decimal_digits(number):
    """
    Count the number of decimal digits in a number
    :param number: The number
    :return: The number of decimal digits
    """
    number_parts = repr(number).split(".")
    if len(number_parts) == 1:
        return 0
    else:
        if number_parts[1] == "0":
            return 0
        else:
            return len(number_parts[1].rstrip("0"))


def get_variable_bounds(m, data, basis):
    coeff_lb_count = defaultdict(int)
    coeff_ub_count = defaultdict(int)

    min_exponent = float("inf")
    max_exponent = -float("inf")

    for var in m.getVars():
        if var.LB != 0 and abs(var.LB) < gp.GRB.INFINITY:
            exponent = int(math.floor(math.log(abs(var.LB), basis)))
            coeff_lb_count[exponent] += 1
            if exponent < min_exponent:
                min_exponent = exponent
            if exponent > max_exponent:
                max_exponent = exponent

        if var.UB != 0 and abs(var.UB) < gp.GRB.INFINITY:
            exponent = int(math.floor(math.log(abs(var.UB), basis)))
            coeff_ub_count[exponent] += 1
            if exponent < min_exponent:
                min_exponent = exponent
            if exponent > max_exponent:
                max_exponent = exponent

    if min_exponent == float("inf"):
        return [], []

    else:
        coeff_lb_result = [
            [exponent, coeff_lb_count[exponent]]
            for exponent in range(min_exponent, max_exponent + 1)
        ]
        coeff_ub_result = [
            [exponent, coeff_ub_count[exponent]]
            for exponent in range(min_exponent, max_exponent + 1)
        ]
        return coeff_lb_result, coeff_ub_result


def get_vector_frequencies(vec, base=10):
    count = defaultdict(int)
    result = []

    for v in vec:
        if v != 0 and abs(v) != float("inf"):
            count[int(math.floor(math.log(abs(v), base)))] += 1

    if len(count) > 0:
        result = [
            [exponent, count[exponent]]
            for exponent in range(min(count.keys()), max(count.keys()) + 1)
        ]

    return result


def get_obj_c_frequencies(model, base=10):
    int_result = []
    cont_result = []
    intobjcoeffs = []
    contobjcoeffs = []
    for x in model.getVars():
        if x.Obj != 0:
            if x.VType == "C":
                contobjcoeffs.append(x.Obj)
            else:
                intobjcoeffs.append(x.Obj)

    if len(contobjcoeffs) > 0:
        cont_result = get_vector_frequencies(contobjcoeffs, base)
    if len(intobjcoeffs) > 0:
        int_result = get_vector_frequencies(intobjcoeffs, base)

    return cont_result, int_result


#
#   Extract linear constraints with names specified by prefix and optional
#   suffix
#


def extractconstraints(model, prefix, suffix=None):
    cons = []
    for c in model.getConstrs():
        cstr = c.ConstrName
        if prefix != None:
            if not cstr.startswith(prefix):
                continue
        if suffix != None:
            if not cstr.endswith(suffix):
                continue
        cons.append(c)

    return cons


#
#   Extract quadratic constraints with names specified by prefix
#   and optional suffix
#


def extractqconstraints(model, prefix, suffix=None):
    qcons = []
    for qc in model.getQConstrs():
        cstr = qc.QCName
        if prefix != None:
            if not cstr.startswith(prefix):
                continue
        if suffix != None:
            if not cstr.endswith(suffix):
                continue
        qcons.append(qc)

    return qcons


#
#   Extract variables with names specified by prefix and optional suffix
#   Also allow for extracting by upper or lower bounds above/below a
#   specified threshold.  Also allow for extracting by a string of types
#   consisting of a subset of "BIC" (Binary/Integer/Continuous)
#


def extractvariables(model, prefix=None, suffix=None, bndthresh=None, types=None):
    vars = []
    for v in model.getVars():
        vstr = v.VarName
        if prefix != None:
            if not vstr.startswith(prefix):
                continue
        if suffix != None:
            if not vstr.endswith(suffix):
                continue
        if bndthresh != None:
            if (v.UB >= GRB.INFINITY or v.UB < bndthresh) and (
                v.LB <= -GRB.INFINITY or v.LB > -bndthresh
            ):
                continue
        if types != None:
            hit = False
            for char in types:
                if v.Vtype == char:
                    hit = True
                    break
            if not hit:
                continue
        vars.append(v)

    return vars


#
#   Returns the variables contained in a linear expression
#
def extractvariablesfromexpr(linexpr):
    retvars = []
    for j in range(linexpr.size()):
        retvars.append(linexpr.getVar(j))

    return retvars


#
#   Iterate through the dictionary in data, which is assumed
#   to contain string keys and  integer values, and print the results
#
def printvals(datadict):
    for datakey in datadict:
        print(datakey, " = ", str(datadict[datakey]))
